# Copyright (c) 2016, Science and Technology Facilities Council
# This software is distributed under a BSD licence. See LICENSE.txt.

"""Tests for the mrcfile package."""

# Import Python 3 features for future-proofing
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import unittest

from .test_bzip2mrcfile import Bzip2MrcFileTest
from .test_command_line import CommandLineTest
from .test_dtypes import DtypesTest
from .test_future_mrcfile import FutureMrcFileTest
from .test_gzipmrcfile import GzipMrcFileTest
from .test_load_functions import LoadFunctionTest, LoadFunctionTestWithPathlib
from .test_mrcfile import MrcFileTest
from .test_mrcinterpreter import MrcInterpreterTest
from .test_mrcmemmap import MrcMemmapTest
from .test_mrcobject import MrcObjectTest
from .test_utils import UtilsTest
from .test_validation import ValidationTest

test_classes = [
    Bzip2MrcFileTest,
    CommandLineTest,
    DtypesTest,
    FutureMrcFileTest,
    GzipMrcFileTest,
    LoadFunctionTest,
    LoadFunctionTestWithPathlib,
    MrcFileTest,
    MrcInterpreterTest,
    MrcObjectTest,
    MrcMemmapTest,
    UtilsTest,
    ValidationTest
]

def load_tests(loader, tests, pattern):
    suite = unittest.TestSuite()
    for test_class in test_classes:
        suite.addTest(loader.loadTestsFromTestCase(test_class))
    return suite



# Copyright (c) 2022, Science and Technology Facilities Council
# This software is distributed under a BSD licence. See LICENSE.txt.

"""
Tests for dtypes.py
"""

# Import Python 3 features for future-proofing
# Deliberately do NOT import unicode_literals due to a bug in numpy dtypes:
# https://github.com/numpy/numpy/issues/2407
from __future__ import absolute_import, division, print_function

import unittest

import mrcfile.dtypes as dtypes
import mrcfile.utils as utils
from .helpers import AssertRaisesRegexMixin


class DtypesTest(AssertRaisesRegexMixin, unittest.TestCase):
    
    """Unit tests for mrcfile.dtypes"""

    def test_invalid_byte_order_raises_exception(self):
        with self.assertRaisesRegex(ValueError, "Unrecognised byte order indicator"):
            _ = dtypes.get_ext_header_dtype('', 'a')
    
    def test_fei1_ext_header_with_native_byte_order(self):
        dtype = dtypes.get_ext_header_dtype(b'FEI1')
        assert dtype.itemsize == 768
        assert utils.byte_orders_equal(dtype['Metadata size'].byteorder, '=')
        with self.assertRaises(KeyError):
            _ = dtype['Scan rotation']
        # Bitmasks should always be little-endian
        assert utils.byte_orders_equal(dtype['Bitmask 1'].byteorder, '<')
        assert utils.byte_orders_equal(dtype['Bitmask 2'].byteorder, '<')
        assert utils.byte_orders_equal(dtype['Bitmask 3'].byteorder, '<')
        assert utils.byte_orders_equal(dtype['Bitmask 4'].byteorder, '<')

    def test_fei2_ext_header_with_native_byte_order(self):
        dtype = dtypes.get_ext_header_dtype(b'FEI2')
        assert dtype.itemsize == 888
        assert utils.byte_orders_equal(dtype['Metadata size'].byteorder, '=')
        assert dtype['Scan rotation'] is not None
        assert utils.byte_orders_equal(dtype['Scan rotation'].byteorder, '=')
        # Bitmasks should always be little-endian
        assert utils.byte_orders_equal(dtype['Bitmask 1'].byteorder, '<')
        assert utils.byte_orders_equal(dtype['Bitmask 2'].byteorder, '<')
        assert utils.byte_orders_equal(dtype['Bitmask 3'].byteorder, '<')
        assert utils.byte_orders_equal(dtype['Bitmask 4'].byteorder, '<')

    def test_fei1_ext_header_with_little_endian_byte_order(self):
        dtype = dtypes.get_ext_header_dtype(b'FEI1', '<')
        # Normal fields should match the requested byte order
        assert utils.byte_orders_equal(dtype['Metadata size'].byteorder, '<')
        assert utils.byte_orders_equal(dtype['Alpha tilt'].byteorder, '<')
        # Bitmasks should always be little-endian
        assert utils.byte_orders_equal(dtype['Bitmask 1'].byteorder, '<')
        assert utils.byte_orders_equal(dtype['Bitmask 2'].byteorder, '<')
        assert utils.byte_orders_equal(dtype['Bitmask 3'].byteorder, '<')
        assert utils.byte_orders_equal(dtype['Bitmask 4'].byteorder, '<')

    def test_fei2_ext_header_with_little_endian_byte_order(self):
        dtype = dtypes.get_ext_header_dtype(b'FEI2', '<')
        # Normal fields should match the requested byte order
        assert utils.byte_orders_equal(dtype['Metadata size'].byteorder, '<')
        assert utils.byte_orders_equal(dtype['Alpha tilt'].byteorder, '<')
        assert utils.byte_orders_equal(dtype['Scan rotation'].byteorder, '<')
        # Bitmasks should always be little-endian
        assert utils.byte_orders_equal(dtype['Bitmask 1'].byteorder, '<')
        assert utils.byte_orders_equal(dtype['Bitmask 2'].byteorder, '<')
        assert utils.byte_orders_equal(dtype['Bitmask 3'].byteorder, '<')
        assert utils.byte_orders_equal(dtype['Bitmask 4'].byteorder, '<')

    def test_fei1_ext_header_with_big_endian_byte_order(self):
        dtype = dtypes.get_ext_header_dtype(b'FEI1', '>')
        # Normal fields should match the requested byte order
        assert utils.byte_orders_equal(dtype['Metadata size'].byteorder, '>')
        assert utils.byte_orders_equal(dtype['Alpha tilt'].byteorder, '>')
        # Bitmasks should always be little-endian
        assert utils.byte_orders_equal(dtype['Bitmask 1'].byteorder, '<')
        assert utils.byte_orders_equal(dtype['Bitmask 2'].byteorder, '<')
        assert utils.byte_orders_equal(dtype['Bitmask 3'].byteorder, '<')
        assert utils.byte_orders_equal(dtype['Bitmask 4'].byteorder, '<')

    def test_fei2_ext_header_with_big_endian_byte_order(self):
        dtype = dtypes.get_ext_header_dtype(b'FEI2', '>')
        # Normal fields should match the requested byte order
        assert utils.byte_orders_equal(dtype['Metadata size'].byteorder, '>')
        assert utils.byte_orders_equal(dtype['Alpha tilt'].byteorder, '>')
        assert utils.byte_orders_equal(dtype['Scan rotation'].byteorder, '>')
        # Bitmasks should always be little-endian
        assert utils.byte_orders_equal(dtype['Bitmask 1'].byteorder, '<')
        assert utils.byte_orders_equal(dtype['Bitmask 2'].byteorder, '<')
        assert utils.byte_orders_equal(dtype['Bitmask 3'].byteorder, '<')
        assert utils.byte_orders_equal(dtype['Bitmask 4'].byteorder, '<')


if __name__ == '__main__':
    unittest.main()



# coding: utf-8
# Copyright (c) 2016, Science and Technology Facilities Council
# This software is distributed under a BSD licence. See LICENSE.txt.

"""
Tests for utils.py
"""

# Import Python 3 features for future-proofing
# Deliberately do NOT import unicode_literals due to a bug in numpy dtypes:
# https://github.com/numpy/numpy/issues/2407
# and also because some tests for string <-> byte conversion need to test both unicode
# and non-unicode literals.
from __future__ import absolute_import, division, print_function

import sys
import unittest

import numpy as np

import mrcfile.utils as utils
from .helpers import AssertRaisesRegexMixin
from mrcfile.dtypes import HEADER_DTYPE


class UtilsTest(AssertRaisesRegexMixin, unittest.TestCase):
    
    """Unit tests for mrcfile.utils"""
    
    def test_header_dtype_is_correct_length(self):
        assert HEADER_DTYPE.itemsize == 1024
    
    def test_mode_0_is_converted_to_int8(self):
        dtype = utils.dtype_from_mode(0)
        assert dtype == np.dtype(np.int8)
    
    def test_mode_1_is_converted_to_int16(self):
        dtype = utils.dtype_from_mode(1)
        assert dtype == np.dtype(np.int16)
    
    def test_mode_2_is_converted_to_float32(self):
        dtype = utils.dtype_from_mode(2)
        assert dtype == np.dtype(np.float32)
    
    def test_mode_3_raises_exception(self):
        with self.assertRaises(ValueError):
            utils.dtype_from_mode(3)
    
    def test_mode_4_is_converted_to_complex64(self):
        dtype = utils.dtype_from_mode(4)
        assert dtype == np.dtype(np.complex64)
    
    def test_mode_6_is_converted_to_uint16(self):
        dtype = utils.dtype_from_mode(6)
        assert dtype == np.dtype(np.uint16)
    
    def test_undefined_modes_raise_exception(self):
        for mode in (x for x in range(-33, 34, 1) if x not in [0, 1, 2, 4, 6, 12]):
            with self.assertRaises(ValueError):
                utils.dtype_from_mode(mode)
    
    def test_mode_scalar_is_converted_without_error(self):
        dtype = utils.dtype_from_mode(np.float32(1))
        assert dtype == np.dtype(np.int16)
    
    def test_mode_zerodim_array_is_converted_without_error(self):
        dtype = utils.dtype_from_mode(np.array(1))
        assert dtype == np.dtype(np.int16)
    
    def test_mode_onedim_array_is_converted_without_error(self):
        dtype = utils.dtype_from_mode(np.array([1]))
        assert dtype == np.dtype(np.int16)
    
    def test_empty_mode_array_raises_error(self):
        with self.assertRaises(ValueError):
            utils.dtype_from_mode(np.array([]))
    
    def test_mode_array_with_two_items_raises_error(self):
        with self.assertRaises(ValueError):
            utils.dtype_from_mode(np.array([1, 2]))
    
    def test_float16_dtype_is_converted_to_mode_12(self):
        mode = utils.mode_from_dtype(np.dtype(np.float16))
        assert mode == 12
    
    def test_float32_dtype_is_converted_to_mode_2(self):
        mode = utils.mode_from_dtype(np.dtype(np.float32))
        assert mode == 2
    
    def test_float64_dtype_raises_exception(self):
        with self.assertRaises(ValueError):
            utils.mode_from_dtype(np.dtype(np.float64))
    
    def test_float128_dtype_raises_exception(self):
        # float128 only exists on some platforms
        if hasattr(np, 'float128'):
            with self.assertRaises(ValueError):
                utils.mode_from_dtype(np.dtype(np.float128))
    
    def test_int8_dtype_is_converted_to_mode_0(self):
        mode = utils.mode_from_dtype(np.dtype(np.int8))
        assert mode == 0
    
    def test_int16_dtype_is_converted_to_mode_1(self):
        mode = utils.mode_from_dtype(np.dtype(np.int16))
        assert mode == 1
    
    def test_int32_dtype_raises_exception(self):
        with self.assertRaises(ValueError):
            utils.mode_from_dtype(np.dtype(np.int32))
    
    def test_int64_dtype_raises_exception(self):
        with self.assertRaises(ValueError):
            utils.mode_from_dtype(np.dtype(np.int64))
    
    def test_uint8_dtype_is_converted_to_mode_6(self):
        mode = utils.mode_from_dtype(np.dtype(np.uint8))
        assert mode == 6
    
    def test_uint16_dtype_is_converted_to_mode_6(self):
        mode = utils.mode_from_dtype(np.dtype(np.uint16))
        assert mode == 6
    
    def test_uint32_dtype_raises_exception(self):
        with self.assertRaises(ValueError):
            utils.mode_from_dtype(np.dtype(np.uint32))
    
    def test_uint64_dtype_raises_exception(self):
        with self.assertRaises(ValueError):
            utils.mode_from_dtype(np.dtype(np.uint64))
    
    def test_complex64_dtype_is_converted_to_mode_4(self):
        mode = utils.mode_from_dtype(np.dtype(np.complex64))
        assert mode == 4
    
    def test_complex128_dtype_raises_exception(self):
        with self.assertRaises(ValueError):
            utils.mode_from_dtype(np.dtype(np.complex128))
    
    def test_string_dtype_raises_exception(self):
        with self.assertRaises(ValueError):
            utils.mode_from_dtype(np.dtype('S1'))
    
    def test_unicode_dtype_raises_exception(self):
        with self.assertRaises(ValueError):
            utils.mode_from_dtype(np.dtype('U1'))
    
    def test_bool_dtype_raises_exception(self):
        with self.assertRaises(ValueError):
            utils.mode_from_dtype(np.dtype(bool))
    
    def test_object_dtype_raises_exception(self):
        with self.assertRaises(ValueError):
            utils.mode_from_dtype(np.dtype(object))
    
    def test_structured_dtype_raises_exception(self):
        with self.assertRaises(ValueError):
            utils.mode_from_dtype(np.dtype([('f1', np.int32)]))
    
    def test_little_endian_machine_stamp(self):
        machst = utils.machine_stamp_from_byte_order('<')
        assert machst == bytearray((0x44, 0x44, 0x00, 0x00))
    
    def test_big_endian_machine_stamp(self):
        machst = utils.machine_stamp_from_byte_order('>')
        assert machst == bytearray((0x11, 0x11, 0x00, 0x00))
    
    def test_native_machine_stamp(self):
        machst = utils.machine_stamp_from_byte_order()
        if sys.byteorder == 'little':
            assert machst == utils.machine_stamp_from_byte_order('<')
        else:
            assert machst == utils.machine_stamp_from_byte_order('>')
    
    def test_normalise_little_endian_byte_order(self):
        assert utils.normalise_byte_order('<') == '<'
    
    def test_normalise_big_endian_byte_order(self):
        assert utils.normalise_byte_order('>') == '>'
    
    def test_normalise_native_byte_order(self):
        if sys.byteorder == 'little':
            assert utils.normalise_byte_order('=') == '<'
        else:
            assert utils.normalise_byte_order('=') == '>'
    
    def test_normalise_unknown_byte_orders(self):
        for byte_order in ['|', 'I', 'other', 'S', 'N', 'L', 'B']:
            with self.assertRaisesRegex(ValueError,
                                        "Unrecognised byte order indicator"):
                utils.normalise_byte_order(byte_order)
    
    def test_native_byte_orders_equal(self):
        assert utils.byte_orders_equal('=', '=')
    
    def test_little_byte_order_equals_native(self):
        assert utils.byte_orders_equal('<', '=') == (sys.byteorder == 'little')
    
    def test_big_byte_order_equals_native(self):
        assert utils.byte_orders_equal('>', '=') == (sys.byteorder == 'big')
    
    def test_little_byte_orders_equal(self):
        assert utils.byte_orders_equal('<', '<')
    
    def test_big_byte_orders_equal(self):
        assert utils.byte_orders_equal('>', '>')
    
    def test_unequal_byte_orders(self):
        assert not utils.byte_orders_equal('>', '<')
    
    def test_equality_of_invalid_byte_orders(self):
        for pair in [('|', '<'),
                     ('|', '>'),
                     ('|', '='),
                     ('<', '|'),
                     ('>', '|'),
                     ('=', '|'),
                     ('|', '|')]:
            with self.assertRaisesRegex(ValueError,
                                        "Unrecognised byte order indicator"):
                utils.byte_orders_equal(*pair)
    
    def test_unknown_byte_order_raises_exception(self):
        with self.assertRaisesRegex(ValueError, "Unrecognised byte order indicator"):
            utils.machine_stamp_from_byte_order('|')
    
    def test_spacegroup_is_volume_stack(self):
        for ispg in range(-2000, 2000):
            assert utils.spacegroup_is_volume_stack(ispg) == (401 <= ispg <= 630)

    def test_pretty_machine_stamp(self):
        machst = utils.machine_stamp_from_byte_order('<')
        assert utils.pretty_machine_stamp(machst) == "0x44 0x44 0x00 0x00"

    def test_is_printable_ascii_with_printable_string(self):
        assert utils.is_printable_ascii('Letters + digits012359 + punctuation*&^%$!#')

    def test_is_printable_ascii_rejects_newline(self):
        assert not utils.is_printable_ascii('\n')

    def test_is_printable_ascii_rejects_tab(self):
        assert not utils.is_printable_ascii('\t')

    def test_is_printable_ascii_rejects_null_byte(self):
        assert not utils.is_printable_ascii('\x00')

    def test_is_printable_ascii_rejects_other_unprintable_bytes(self):
        # Try a few others; not comprehensive.
        assert not utils.is_printable_ascii('\x01')
        assert not utils.is_printable_ascii('\x02')
        assert not utils.is_printable_ascii('\x03')

    def test_is_printable_ascii_rejects_non_ascii_printable_character(self):
        assert not utils.is_printable_ascii('£')

    def test_printable_string_from_bytes_leaves_ascii_unchanged(self):
        val = b'Letters + digits012359 + punctuation*&^%$!#'
        result = utils.printable_string_from_bytes(val)
        assert result == 'Letters + digits012359 + punctuation*&^%$!#'

    def test_printable_string_from_bytes_strips_unprintable_characters(self):
        val = b'Test \n non-printable string \t \x01\x00'
        assert utils.printable_string_from_bytes(val) == 'Test  non-printable string  '

    def test_printable_string_from_bytes_strips_non_ascii_characters(self):
        val = b'Test non-ASCII string \xa3'
        assert utils.printable_string_from_bytes(val) == 'Test non-ASCII string '

    def test_bytes_from_string_with_ascii_string(self):
        assert utils.bytes_from_string('ASCII string 123#!') == b'ASCII string 123#!'

    def test_bytes_from_string_with_non_ascii_string_raises_exception(self):
        with self.assertRaises(UnicodeError):
            utils.bytes_from_string('Non-ASCII string £')

    def test_bytes_from_string_with_unicode_string(self):
        # To make sure things work properly in Python 2
        assert utils.bytes_from_string(u'Unicode string') == b'Unicode string'

    def test_bytes_from_string_with_unicode_non_ascii_string_raises_exception(self):
        # To make sure things work properly in Python 2
        with self.assertRaises(UnicodeError):
            assert utils.bytes_from_string(u'Unicode non-ASCII £')


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2018, Science and Technology Facilities Council
# This software is distributed under a BSD licence. See LICENSE.txt.

"""
Tests for future_mrcfile.py.
"""

# Import Python 3 features for future-proofing
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import time
import unittest

from mrcfile.future_mrcfile import FutureMrcFile
from . import helpers


def sleep_and_return_args(*args, **kwargs):
    """Simple test function to wait a short time and return its arguments."""
    time.sleep(0.01)
    return [args, kwargs]


class FutureMrcFileTest(helpers.AssertRaisesRegexMixin, unittest.TestCase):

    """Unit tests for FutureMrcFile.

    These tests only make sure that the FutureMrcFile class works correctly
    in terms of function calling and behaviour. Actual testing of asynchronous
    MRC file opening is done in test_load_functions.py.

    """

    def setUp(self):
        self.future_mrc_file = FutureMrcFile(sleep_and_return_args)

    def tearDown(self):
        pass

    def test_cancel(self):
        assert self.future_mrc_file.cancel() == False

    def test_cancelled(self):
        assert self.future_mrc_file.cancelled() == False

    def test_add_done_callback(self):
        with self.assertRaises(NotImplementedError):
            self.future_mrc_file.add_done_callback(sleep_and_return_args)

    def test_running_and_done_status(self):
        assert self.future_mrc_file.running() == True
        assert self.future_mrc_file.done() == False
        self.future_mrc_file.result()
        assert self.future_mrc_file.running() == False
        assert self.future_mrc_file.done() == True

    def test_arguments_passed_correctly(self):
        args = (1, 2, 'a')
        kwargs = {'a': 'b', 'c': 3}
        future_mrc_file = FutureMrcFile(sleep_and_return_args, args, kwargs)
        result = future_mrc_file.result()
        assert result[0] == args
        assert result[1] == kwargs

        # Also check the results from the Future from setUp()
        result_no_args = self.future_mrc_file.result()
        assert result_no_args[0] == ()
        assert result_no_args[1] == {}

    def test_timeout_from_result(self):
        with self.assertRaisesRegex(RuntimeError, "Timed out"):
            FutureMrcFile(sleep_and_return_args).result(0.001)

    def test_timeout_from_exception(self):
        with self.assertRaisesRegex(RuntimeError, "Timed out"):
            FutureMrcFile(sleep_and_return_args).exception(0.001)

    def test_exception(self):
        future_mrc_file = FutureMrcFile(lambda: 1/0)
        ex = future_mrc_file.exception()
        assert isinstance(ex, ZeroDivisionError)
        with self.assertRaises(ZeroDivisionError) as cm:
            future_mrc_file.result()
            assert cm.exception == ex



# Copyright (c) 2016, Science and Technology Facilities Council
# This software is distributed under a BSD licence. See LICENSE.txt.
"""
mrcobject
---------

Module which exports the :class:`MrcObject` class.

Classes:
    :class:`MrcObject`: An object representing image or volume data in the MRC
    format.

"""

# Import Python 3 features for future-proofing
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

from datetime import datetime
import warnings

import numpy as np

from . import utils
from .dtypes import (HEADER_DTYPE, VOXEL_SIZE_DTYPE, NSTART_DTYPE,
                     get_ext_header_dtype)
from .constants import (MAP_ID, IMAGE_STACK_SPACEGROUP, VOLUME_SPACEGROUP,
                        VOLUME_STACK_SPACEGROUP)


class MrcObject(object):

    """An object representing image or volume data in the MRC format.

    The header, extended header and data are stored as numpy arrays and
    exposed as read-only attributes. To replace the data or extended header,
    call :meth:`set_data` or :meth:`set_extended_header`. The header cannot be
    replaced but can be modified in place.

    Voxel size is exposed as a writeable attribute, but is calculated
    on-the-fly from the header's ``cella`` and ``mx``/``my``/``mz`` fields.

    Three-dimensional data can represent either a stack of 2D images, or a 3D
    volume. This is indicated by the header's ``ispg`` (space group) field,
    which is set to 0 for image data and >= 1 for volume data. The
    :meth:`is_single_image`, :meth:`is_image_stack`, :meth:`is_volume` and
    :meth:`is_volume_stack` methods can be used to identify the type of
    information stored in the data array. For 3D data, the
    :meth:`set_image_stack` and :meth:`set_volume` methods can be used to
    switch between image stack and volume interpretations of the data.

    If the data contents have been changed, you can use the
    :meth:`update_header_from_data` and :meth:`update_header_stats` methods to
    make the header consistent with the data. These methods are called
    automatically if the data array is replaced by calling :meth:`set_data`.
    :meth:`update_header_from_data` is fast, even with very large data arrays,
    because it only examines the shape and type of the data array.
    :meth:`update_header_stats` calculates statistics from all items in the
    data array and so can be slow for very large arrays. If necessary, the
    :meth:`reset_header_stats` method can be called to set the header fields to
    indicate that the statistics are undetermined.

    Attributes:

    * :attr:`header`
    * :attr:`extended_header`
    * :attr:`indexed_extended_header`
    * :attr:`data`
    * :attr:`voxel_size`
    * :attr:`nstart`

    Methods:

    * :meth:`set_extended_header`
    * :meth:`set_data`
    * :meth:`is_single_image`
    * :meth:`is_image_stack`
    * :meth:`is_volume`
    * :meth:`is_volume_stack`
    * :meth:`set_image_stack`
    * :meth:`set_volume`
    * :meth:`update_header_from_data`
    * :meth:`update_header_stats`
    * :meth:`reset_header_stats`
    * :meth:`print_header`
    * :meth:`get_labels`
    * :meth:`add_label`

    Attributes and methods relevant to subclasses:

    * ``_read_only``
    * :meth:`_check_writeable`
    * :meth:`_create_default_attributes`
    * :meth:`_close_data`
    * :meth:`_set_new_data`

    """

    def __init__(self, **kwargs):
        """Initialise a new :class:`MrcObject`.

        This initialiser deliberately avoids creating any arrays and simply
        sets the header, extended header and data attributes to :data:`None`.
        This allows subclasses to call :meth:`__init__` at the start of their
        initialisers and then set the attributes themselves, probably by
        reading from a file, or by calling :meth:`_create_default_attributes`
        for a new empty object.

        Note that this behaviour might change in future: this initialiser could
        take optional arguments to allow the header and data to be provided
        by the caller, or might create the standard empty defaults rather than
        setting the attributes to :data:`None`.
        """
        super(MrcObject, self).__init__(**kwargs)

        # Set empty default attributes
        self._header = None
        self._extended_header = None
        self._data = None
        self._read_only = False

    def _check_writeable(self):
        """Check that this MRC object is writeable.

        Raises:
            :exc:`ValueError`: If this object is read-only.
        """
        if self._read_only:
            raise ValueError('MRC object is read-only')

    def _create_default_attributes(self):
        """Set valid default values for the header and data attributes."""
        self._create_default_header()
        self._extended_header = np.empty(0, dtype='V1')
        self._set_new_data(np.empty(0, dtype=np.int8))

    def _create_default_header(self):
        """Create a default MRC file header.

        The header is initialised with standard file type and version
        information, default values for some essential fields, and zeros
        elsewhere. The first text label is also set to indicate the file was
        created by this module.
        """
        self._header = np.zeros(shape=(), dtype=HEADER_DTYPE).view(np.recarray)
        header = self._header
        header.map = MAP_ID
        header.nversion = 20141  # current MRC 2014 format version
        header.machst = utils.machine_stamp_from_byte_order(header.mode.dtype.byteorder)

        # Default space group is P1
        header.ispg = VOLUME_SPACEGROUP

        # Standard cell angles all 90.0 degrees
        default_cell_angle = 90.0
        header.cellb.alpha = default_cell_angle
        header.cellb.beta = default_cell_angle
        header.cellb.gamma = default_cell_angle
        # (this can also be achieved by assigning a 3-tuple to header.cellb
        # directly but using the sub-fields individually is easier to read and
        # understand)

        # Standard axes: columns = X, rows = Y, sections = Z
        header.mapc = 1
        header.mapr = 2
        header.maps = 3

        time = datetime.now().strftime('%Y-%m-%d %H:%M:%S')
        header.label[0] = '{0:40s}{1:>39s} '.format('Created by mrcfile.py',
                                                    time)
        header.nlabl = 1

        self.reset_header_stats()

    @property
    def header(self):
        """Get the header as a :class:`numpy record array <numpy.recarray>`."""
        return self._header

    @property
    def extended_header(self):
        """Get the extended header as a :class:`numpy array <numpy.ndarray>`.

        The dtype will be void (raw data, dtype ``V'``). If the actual data type
        of the extended header is known, the dtype of the array can be changed
        to match. For supported types (e.g. ``'FEI1'`` and ``'FEI2'``), the
        indexed part of the extended header (excluding any zero padding) can be
        accessed using :meth:`indexed_extended_header`.

        The extended header may be modified in place. To replace it completely,
        call :meth:`set_extended_header`.
        """
        return self._extended_header

    @property
    def indexed_extended_header(self):
        """Get the indexed part of the extended header as a
        :class:`numpy array <numpy.ndarray>` with the appropriate dtype set.

        Currently only ``'FEI1'`` and ``'FEI2'`` extended headers are supported.
        Modifications to the indexed extended header will not change the
        extended header data recorded in this :class:`MrcObject`. If the
        extended header type is unrecognised or extended header data is not of
        sufficient length a warning will be produced and the indexed extended
        header will be None.
        """
        # Use the header's byte order for the extended header
        dtype = get_ext_header_dtype(self.header.exttyp,
                                     self.header.mode.dtype.byteorder)

        # Interpret one element
        try:
            if self.extended_header.nbytes < dtype.itemsize:
                raise ValueError
            first = self.extended_header[0:dtype.itemsize]
            first.dtype = dtype
            if first["Metadata size"][0] != dtype.itemsize:
                raise ValueError
        except ValueError:
            warnings.warn("The header has exttyp '{}' but the extended header "
                            "cannot be interpreted as that type"
                            .format(self.header.exttyp), RuntimeWarning)
            return None

        nbytes = int(self.header["nz"]) * dtype.itemsize
        try:
            if self.extended_header.nbytes < nbytes:
                raise ValueError
            full = self.extended_header[0:nbytes]
            full.dtype = dtype
        except ValueError:
            warnings.warn("The header has exttyp '{}' but the extended header "
                            "cannot be interpreted as that type"
                            .format(self.header.exttyp), RuntimeWarning)
            return None

        return full

    def set_extended_header(self, extended_header):
        """Replace the extended header.

        If you set the extended header you should also set the
        ``header.exttyp`` field to indicate the type of extended header.

        Raises:
            :exc:`ValueError`: If the new extended header has more than
                2,147,483,647 bytes (and therefore its size cannot be stored
                in the header).
        """
        self._check_writeable()
        if extended_header.nbytes > np.iinfo(np.int32).max:
            raise ValueError("New extended header is too large! It has {} "
                             "bytes. The maximum allowed is {}."
                             .format(extended_header.nbytes,
                                     np.iinfo(np.int32).max))
        self._extended_header = extended_header
        self.header.nsymbt = extended_header.nbytes

    @property
    def data(self):
        """Get the data as a :class:`numpy array <numpy.ndarray>`."""
        return self._data

    def set_data(self, data):
        """Replace the data array.

        This replaces the current data with the given array (or a copy of it),
        and updates the header to match the new data dimensions. The data
        statistics (min, max, mean and rms) stored in the header will also be
        updated.

        Raises:
            :exc:`ValueError`: if the new data has a dimension larger than
                2,147,483,647 (and therefore its size cannot be stored in the
                header).

        Warns:
            RuntimeWarning: If the data array contains Inf or NaN values.
        """
        self._check_writeable()

        # Check if the new data's dtype is valid without changes
        mode = utils.mode_from_dtype(data.dtype)
        new_dtype = (utils.dtype_from_mode(mode)
                     .newbyteorder(data.dtype.byteorder))
        
        for dim in data.shape:
            if dim > np.iinfo(np.int32).max:
                raise ValueError("New data array is too large! Found a "
                                 "dimension of size {}. The maximum allowed "
                                 "is {}."
                                 .format(dim, np.iinfo(np.int32).max))
        
        # Set new_dtype to None if it is the same type as the original array,
        # to avoid numpy >= 1.24 copying the array unnecessarily
        if new_dtype == data.dtype:
            new_dtype = None

        # Copy the data if necessary to ensure correct dtype and C ordering
        new_data = np.ascontiguousarray(data, dtype=new_dtype)

        # Replace the old data array with the new one, and update the header
        self._close_data()
        self._set_new_data(new_data)
        self.update_header_from_data()
        self.update_header_stats()

    def _close_data(self):
        """Close the data array."""
        self._data = None

    def _set_new_data(self, data):
        """Replace the data array with a new one.

        The new data array is not checked - it must already be valid for use in
        an MRC file.
        """
        self._data = data

    @property
    def voxel_size(self):
        """Get or set the voxel size in angstroms.

        The voxel size is returned as a structured NumPy :class:`record array
        <numpy.recarray>` with three fields (x, y and z). For example:

        >>> mrc.voxel_size
        rec.array((0.44825, 0.3925, 0.45874998),
          dtype=[('x', '<f4'), ('y', '<f4'), ('z', '<f4')])
        >>> mrc.voxel_size.x
        array(0.44825, dtype=float32)

        Note that changing the voxel_size array in-place will *not* change the
        voxel size in the file -- to prevent this being overlooked
        accidentally, the writeable flag is set to :data:`False` on the
        voxel_size array.

        To set the voxel size, assign a new value to the voxel_size attribute.
        You may give a single number, a 3-tuple ``(x, y ,z)`` or a modified
        version of the voxel_size array. The following examples are all
        equivalent:

        >>> mrc.voxel_size = 1.0

        >>> mrc.voxel_size = (1.0, 1.0, 1.0)

        >>> vox_sizes = mrc.voxel_size
        >>> vox_sizes.flags.writeable = True
        >>> vox_sizes.x = 1.0
        >>> vox_sizes.y = 1.0
        >>> vox_sizes.z = 1.0
        >>> mrc.voxel_size = vox_sizes
        """
        x = self.header.cella.x / self.header.mx
        y = self.header.cella.y / self.header.my
        z = self.header.cella.z / self.header.mz
        sizes = np.rec.array((x, y, z), VOXEL_SIZE_DTYPE)
        sizes.flags.writeable = False
        return sizes

    @voxel_size.setter
    def voxel_size(self, voxel_size):
        self._check_writeable()
        try:
            # First, assume we have a single numeric value
            sizes = (float(voxel_size),) * 3
        except TypeError:
            try:
                # Not a single value. Next, if voxel_size is an array (as
                # produced by the voxel_size getter), item() gives a 3-tuple
                sizes = voxel_size.item()
            except AttributeError:
                # If the item() method doesn't exist, assume we have a 3-tuple
                sizes = voxel_size
        self._set_voxel_size(*sizes)

    def _set_voxel_size(self, x_size, y_size, z_size):
        """Set the voxel size.

        Args:
            x_size: The voxel size in the X direction, in angstroms
            y_size: The voxel size in the Y direction, in angstroms
            z_size: The voxel size in the Z direction, in angstroms
        """
        self.header.cella.x = x_size * self.header.mx
        self.header.cella.y = y_size * self.header.my
        self.header.cella.z = z_size * self.header.mz

    @property
    def nstart(self):
        """Get or set the grid start locations.

        This provides a convenient way to get and set the values of the
        header's ``nxstart``, ``nystart`` and ``nzstart`` fields. Note that
        these fields are integers and are measured in voxels, not angstroms.
        The start locations are returned as a structured NumPy :class:`record
        array <numpy.recarray>` with three fields (x, y and z). For example:

        >>> mrc.header.nxstart
        array(0, dtype=int32)
        >>> mrc.header.nystart
        array(-21, dtype=int32)
        >>> mrc.header.nzstart
        array(-12, dtype=int32)
        >>> mrc.nstart
        rec.array((0, -21, -12),
          dtype=[('x', '<i4'), ('y', '<i4'), ('z', '<i4')])
        >>> mrc.nstart.y
        array(-21, dtype=int32)

        Note that changing the nstart array in-place will *not* change the
        values in the file -- to prevent this being overlooked accidentally,
        the writeable flag is set to :data:`False` on the nstart array.

        To set the start locations, assign a new value to the nstart
        attribute. You may give a single number, a 3-tuple ``(x, y ,z)`` or a
        modified version of the nstart array. The following examples are all
        equivalent:

        >>> mrc.nstart = -150

        >>> mrc.nstart = (-150, -150, -150)

        >>> starts = mrc.nstart
        >>> starts.flags.writeable = True
        >>> starts.x = -150
        >>> starts.y = -150
        >>> starts.z = -150
        >>> mrc.nstart = starts
        """
        x = self.header.nxstart
        y = self.header.nystart
        z = self.header.nzstart
        nstart = np.rec.array((x, y, z), NSTART_DTYPE)
        nstart.flags.writeable = False
        return nstart

    @nstart.setter
    def nstart(self, nstart):
        self._check_writeable()
        try:
            # First, assume we have a single numeric value
            starts = (int(nstart),) * 3
        except TypeError:
            try:
                # Not a single value. Next, if nstart is an array (as
                # produced by the nstart getter), item() gives a 3-tuple
                starts = nstart.item()
            except AttributeError:
                # If the item() method doesn't exist, assume we have a 3-tuple
                starts = nstart
        self._set_nstart(*starts)

    def _set_nstart(self, nxstart, nystart, nzstart):
        """Set the grid start locations.

        Args:
            nxstart: The location of the first column in the unit cell
            nystart: The location of the first row in the unit cell
            nzstart: The location of the first section in the unit cell
        """
        self.header.nxstart = nxstart
        self.header.nystart = nystart
        self.header.nzstart = nzstart

    def is_single_image(self):
        """Identify whether the file represents a single image.

        Returns:
            :data:`True` if the data array is two-dimensional.
        """
        return self.data.ndim == 2

    def is_image_stack(self):
        """Identify whether the file represents a stack of images.

        Returns:
            :data:`True` if the data array is three-dimensional and the space group
            is zero.
        """
        return (self.data.ndim == 3
                and self.header.ispg == IMAGE_STACK_SPACEGROUP)

    def is_volume(self):
        """Identify whether the file represents a volume.

        Returns:
            :data:`True` if the data array is three-dimensional and the space
            group is not zero.
        """
        return (self.data.ndim == 3
                and self.header.ispg != IMAGE_STACK_SPACEGROUP)

    def is_volume_stack(self):
        """Identify whether the file represents a stack of volumes.

        Returns:
            :data:`True` if the data array is four-dimensional.
        """
        return self.data.ndim == 4

    def set_image_stack(self):
        """Change three-dimensional data to represent an image stack.

        This method changes the space group number (``header.ispg``) to zero.

        Raises:
            :exc:`ValueError`: If the data array is not three-dimensional.
        """
        self._check_writeable()
        if self.data.ndim != 3:
            raise ValueError('Only 3D data can be changed into an image stack')
        self.header.ispg = IMAGE_STACK_SPACEGROUP
        self.header.mz = 1

    def set_volume(self):
        """Change three-dimensional data to represent a volume.

        If the space group was previously zero (representing an image stack),
        this method sets it to one. Otherwise the space group is not changed.

        Raises:
            :exc:`ValueError`: If the data array is not three-dimensional.
        """
        self._check_writeable()
        if self.data.ndim != 3:
            raise ValueError('Only 3D data can be changed into a volume')
        if self.is_image_stack():
            self.header.ispg = VOLUME_SPACEGROUP
            self.header.mz = self.header.nz

    def update_header_from_data(self):
        """Update the header from the data array.

        This function updates the header byte order and machine stamp to match
        the byte order of the data. It also updates the file mode, space group
        and the dimension fields ``nx``, ``ny``, ``nz``, ``mx``, ``my`` and
        ``mz``.

        If the data is 2D, the space group is set to 0 (image stack). For 3D
        data the space group is not changed, and for 4D data the space group is
        set to 401 (simple P1 volume stack) unless it is already in the volume
        stack range (401--630).

        This means that new 3D data will be treated as an image stack if the
        previous data was a single image or image stack, or as a volume if the
        previous data was a volume or volume stack.

        Note that this function does *not* update the data statistics fields in
        the header (``dmin``, ``dmax``, ``dmean`` and ``rms``). Use the
        :meth:`update_header_stats` function to update the statistics.
        (This is for performance reasons -- updating the statistics can take a
        long time for large data sets, but updating the other header
        information is always fast because only the type and shape of the data
        array need to be inspected.)
        """
        self._check_writeable()

        # Check the dtype is one we can handle and update mode to match
        header = self.header
        header.mode = utils.mode_from_dtype(self.data.dtype)

        # Ensure header byte order and machine stamp match the data's byte order
        data_byte_order = self.data.dtype.byteorder
        header_byte_order = header.mode.dtype.byteorder
        if (data_byte_order != '|'
            and not utils.byte_orders_equal(data_byte_order, header_byte_order)):
            header.byteswap(True)
            header.dtype = header.dtype.newbyteorder(data_byte_order)
        header.machst = utils.machine_stamp_from_byte_order(header.mode.dtype
                                                            .byteorder)

        shape = self.data.shape
        axes = len(shape)
        if axes == 2:
            # Single image. Space group 0, nz = mz = 1
            header.ispg = IMAGE_STACK_SPACEGROUP
            header.nx = header.mx = shape[1]
            header.ny = header.my = shape[0]
            header.nz = header.mz = 1
        elif axes == 3:
            header.nx = header.mx = shape[2]
            header.ny = header.my = shape[1]
            if header.ispg == IMAGE_STACK_SPACEGROUP:
                # Image stack. mz = 1, nz = sections in the volume
                header.mz = 1
                header.nz = shape[0]
            else:
                # Volume. nz = mz = sections in the volume
                header.nz = header.mz = shape[0]
        elif axes == 4:
            # Volume stack. Space group 401, mz = secs per vol, nz = total sections
            if not utils.spacegroup_is_volume_stack(header.ispg):
                header.ispg = VOLUME_STACK_SPACEGROUP
            header.nx = header.mx = shape[3]
            header.ny = header.my = shape[2]
            header.mz = shape[1]
            header.nz = shape[0] * shape[1]
        else:
            raise ValueError('Data must be 2-, 3- or 4-dimensional')

    def update_header_stats(self):
        """Update the header's ``dmin``, ``dmax``, ``dmean`` and ``rms`` fields
        from the data.

        Note that this can take some time with large files, particularly with
        files larger than the currently available memory.

        Warns:
            RuntimeWarning: If the data array contains Inf or NaN values.
        """
        self._check_writeable()

        if self.data.size > 0:
            # Header stats are always in float32. If we have complex data, this doesn't
            # make sense so just set rms and leave min, max and mean at their default
            # un-set values
            if np.iscomplexobj(self.data):
                # Avoid ComplexWarning by explicitly taking the real part
                self.header.rms = np.float32(self.data.std().real)
            else:
                min = self.data.min()
                max = self.data.max()

                if np.isnan(min):
                    warnings.warn("Data array contains NaN values", RuntimeWarning)
                if np.isinf(min) or np.isinf(max):
                    warnings.warn("Data array contains infinite values", RuntimeWarning)

                self.header.dmin = np.float32(min)
                self.header.dmax = np.float32(max)
                self.header.dmean = self.data.mean(dtype=np.float32)
                self.header.rms = self.data.std(dtype=np.float32)
        else:
            self.reset_header_stats()

    def reset_header_stats(self):
        """Set the header statistics to indicate that the values are unknown."""
        self._check_writeable()

        self.header.dmin = 0
        self.header.dmax = -1
        self.header.dmean = -2
        self.header.rms = -1

    def print_header(self, print_file=None):
        """Print the contents of all header fields.

        Args:
            print_file: The output text stream to use for printing the header.
                This is passed directly to the ``file`` argument of Python's
                :func:`print` function. The default is :data:`None`, which
                means output will be printed to :data:`sys.stdout`.
        """
        for item in self.header.dtype.names:
            print('{0:15s} : {1}'.format(item, self.header[item]),
                  file=print_file)

    def get_labels(self):
        """Get the labels from the MRC header.

        Up to ten labels are stored in the header as arrays of 80 bytes. This method
        returns the labels as Python strings, filtered to remove non-printable
        characters. To access the raw bytes (including any non-printable characters) use
        the ``header.label`` attribute (and note that ``header.nlabl`` stores the number
        of labels currently set).

        Returns:
            The labels, as a list of strings. The list will contain between 0 and 10
            items, each containing up to 80 characters.
        """
        return [
            utils.printable_string_from_bytes(label)
            for label in self.header.label[:self.header.nlabl]
        ]

    def add_label(self, label):
        """Add a label to the MRC header.

        The new label will be stored after any labels already in the header. If all ten
        labels are already in use, an exception will be raised.

        Future versions of this method might add checks to ensure that labels containing
        valid text are not overwritten even if the ``nlabl`` value is incorrect.

        Args:
            label: The label value to store, as a string containing only printable
                ASCII characters.

        Raises:
            :exc:`ValueError`: If the label is longer than 80 bytes or contains
                non-printable or non-ASCII characters.
            :exc:`IndexError`: If the file already contains 10 labels and so an
                additional label cannot be stored.
        """
        if not utils.is_printable_ascii(label):
            raise ValueError("Label contains non-printable or non-ASCII characters")
        label_bytes = utils.bytes_from_string(label)
        if len(label_bytes) > 80:
            raise ValueError("Label value has more than 80 bytes")
        self.header.label[self.header.nlabl] = label
        self.header.nlabl += 1

    def validate(self, print_file=None):
        """Validate this MrcObject.

        This method runs a series of tests to check whether this object
        complies strictly with the MRC2014 format specification:

        #. MRC format ID string: The header's ``map`` field must contain
           "MAP ".
        #. Machine stamp: The machine stamp should contain one of
           ``0x44 0x44 0x00 0x00``, ``0x44 0x41 0x00 0x00`` or
           ``0x11 0x11 0x00 0x00``.
        #. MRC mode: the ``mode`` field should be one of the supported mode
           numbers: 0, 1, 2, 4, 6 or 12. (Note that MRC modes 3 and 101 are
           also valid according to the MRC 2014 specification but are not
           supported by mrcfile.)
        #. Map and cell dimensions: The header fields ``nx``, ``ny``, ``nz``,
           ``mx``, ``my``, ``mz``, ``cella.x``, ``cella.y`` and ``cella.z``
           must all be positive numbers.
        #. Axis mapping: Header fields ``mapc``, ``mapr`` and ``maps`` must
           contain the values 1, 2, and 3 (in any order).
        #. Volume stack dimensions: If the spacegroup is in the range 401--630,
           representing a volume stack, the ``nz`` field should be exactly
           divisible by ``mz`` to represent the number of volumes in the stack.
        #. Header labels: The ``nlabl`` field should be set to indicate the
           number of labels in use, and the labels in use should appear first
           in the label array (that is, there should be no blank labels between
           text-filled ones).
        #. MRC format version: The ``nversion`` field should be 20140 or 20141
           for compliance with the MRC2014 standard.
        #. Extended header type: If an extended header is present, the
           ``exttyp`` field should be set to indicate the type of extended
           header.
        #. Data statistics: The statistics in the header should be correct for
           the actual data, or marked as undetermined.

        Args:
            print_file: The output text stream to use for printing messages
                about the validation. This is passed directly to the ``file``
                argument of Python's :func:`print` function. The default is
                :data:`None`, which means output will be printed to
                :data:`sys.stdout`.

        Returns:
            :data:`True` if this MrcObject  is valid, or :data:`False` if it
            does not meet the MRC format specification in any way.
        """
        valid = True

        def log(message):
            print(message, file=print_file)

        # Check map ID string
        if self.header.map != MAP_ID:
            log("Map ID string is incorrect: found {0}, should be {1}"
                .format(self.header.map, MAP_ID))
            valid = False

        # Check machine stamp
        try:
            utils.byte_order_from_machine_stamp(self.header.machst)
        except ValueError:
            pretty_bytes = utils.pretty_machine_stamp(self.header.machst)
            log("Invalid machine stamp: " + pretty_bytes)
            valid = False

        # Check mode is valid
        try:
            utils.dtype_from_mode(self.header.mode)
        except ValueError:
            log("Invalid mode: {0}".format(self.header.mode))
            valid = False

        # Check map dimensions and other fields are non-negative
        for field in ['nx', 'ny', 'nz', 'mx', 'my', 'mz', 'ispg', 'nlabl']:
            if self.header[field] < 0:
                log("Header field '{0}' is negative".format(field))
                valid = False

        # Check cell dimensions are non-negative
        for field in ['x', 'y', 'z']:
            if self.header.cella[field] < 0:
                log("Cell dimension '{0}' is negative".format(field))
                valid = False

        # Check axis mapping is valid
        axes = set()
        for field in ['mapc', 'mapr', 'maps']:
            axes.add(int(self.header[field]))
        if axes != {1, 2, 3}:
            log("Invalid axis mapping: found {0}, should be [1, 2, 3]"
                .format(sorted(list(axes))))
            valid = False

        # Check mz value for volume stacks
        if utils.spacegroup_is_volume_stack(self.header.ispg):
            if self.header.nz % self.header.mz != 0:
                log("Error in dimensions for volume stack: nz should be "
                    "divisible by mz. Found nz = {0}, mz = {1})"
                    .format(self.header.nz, self.header.mz))
                valid = False

        # Check nlabl is correct
        count = 0
        seen_empty_label = False
        for label in self.header.label:
            if len(label.strip()) > 0:
                count += 1
                if seen_empty_label:
                    log("Error in header labels: empty labels appear between "
                        "text-containing labels")
                    valid = False
            else:
                seen_empty_label = True
        if count != self.header.nlabl:
            log("Error in header labels: nlabl is {0} "
                "but {1} labels contain text".format(self.header.nlabl, count))
            valid = False

        # Check MRC format version
        if self.header.nversion not in (20140, 20141):
            log("File does not declare MRC format version 20140 or 20141: nversion ="
                " {0}".format(self.header.nversion))
            valid = False

        # Check extended header type is set to a known value
        valid_exttypes = [b'CCP4', b'MRCO', b'SERI', b'AGAR', b'FEI1', b'FEI2', b'HDF5']
        if self.header.nsymbt > 0 and self.header.exttyp not in valid_exttypes:
            log("Extended header type is undefined or unrecognised: exttyp = "
                "'{0}'".format(self.header.exttyp.item().decode('ascii')))
            valid = False

        # Check data statistics
        real_rms = real_min = real_max = real_mean = 0
        if self.header.rms >= 0:
            if self.data is not None and len(self.data > 0):
                real_rms = self.data.std()
            if not np.isclose(real_rms, self.header.rms, rtol=0.01):
                log("Data statistics appear to be inaccurate: RMS deviation is {0} but"
                    " the value in the header is {1}".format(real_rms, self.header.rms))
                valid = False
        if self.header.dmin < self.header.dmax:
            if self.data is not None and len(self.data > 0):
                real_min = self.data.min()
                real_max = self.data.max()
            if self.header.dmin != real_min:
                log("Data statistics appear to be inaccurate: minimum is {0} but the"
                    " value in the header is {1}".format(real_min, self.header.dmin))
                valid = False
            if self.header.dmax != real_max:
                log("Data statistics appear to be inaccurate: maximum is {0} but the"
                    " value in the header is {1}".format(real_max, self.header.dmax))
                valid = False
        if self.header.dmean > min(self.header.dmin, self.header.dmax):
            if self.data is not None and len(self.data > 0):
                real_mean = self.data.mean(dtype=np.float64)
            if not np.isclose(real_mean, self.header.dmean, rtol=0.01):
                log("Data statistics appear to be inaccurate: mean is {0} but the"
                    " value in the header is {1}".format(real_mean, self.header.dmean))
                valid = False

        return valid



# Copyright (c) 2018, Science and Technology Facilities Council
# This software is distributed under a BSD licence. See LICENSE.txt.

"""
Tests for mrcfile __init__.py loading functions.
"""

# Import Python 3 features for future-proofing
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import mmap
import os
import shutil
import sys
import tempfile
import unittest

import numpy as np

import mrcfile
from mrcfile.bzip2mrcfile import Bzip2MrcFile
from mrcfile.gzipmrcfile import GzipMrcFile
from . import helpers

# Try to import pathlib if we can
pathlib_unavailable = False
try:
    from pathlib import Path
except ImportError:
    try:
        from pathlib2 import Path
    except ImportError:
        pathlib_unavailable = True


class LoadFunctionTest(helpers.AssertRaisesRegexMixin, unittest.TestCase):
    
    """Unit tests for MRC loading functions.
    
    """
    
    def setUp(self):
        super(LoadFunctionTest, self).setUp()
        
        # Set up test files and names to be used
        self.test_data = helpers.get_test_data_path()
        self.test_output = tempfile.mkdtemp()
        self.temp_mrc_name = os.path.join(self.test_output, 'test_mrcfile.mrc')
        self.temp_gz_mrc_name = self.temp_mrc_name + '.gz'
        self.example_mrc_name = os.path.join(self.test_data, 'EMD-3197.map')
        self.gzip_mrc_name = os.path.join(self.test_data, 'emd_3197.map.gz')
        self.bzip2_mrc_name = os.path.join(self.test_data, 'EMD-3197.map.bz2')
        self.slow_mrc_name = os.path.join(self.test_data, 'fei-extended.mrc.gz')
    
    def tearDown(self):
        if os.path.exists(self.test_output):
            shutil.rmtree(self.test_output)
        super(LoadFunctionTest, self).tearDown()
    
    def test_normal_opening(self):
        with mrcfile.open(self.example_mrc_name) as mrc:
            assert repr(mrc) == ("MrcFile('{0}', mode='r')"
                                 .format(self.example_mrc_name))

    @unittest.skipIf(pathlib_unavailable, "pathlib not available")
    def test_normal_opening_pathlib(self):
        """Single test to ensure pathlib functionality is tested even if there's
        a problem with the LoadFunctionTestWithPathlib class"""
        path = Path(self.example_mrc_name)
        with mrcfile.open(path) as mrc:
            assert repr(mrc) == ("MrcFile('{0}', mode='r')"
                                 .format(self.example_mrc_name))

    def test_read(self):
        volume = mrcfile.read(self.example_mrc_name)
        assert isinstance(volume, np.ndarray)
        assert volume.shape, volume.dtype == ((20, 20, 20), np.float32)
        assert volume.flags.writeable is True
    
    def test_gzip_opening(self):
        with mrcfile.open(self.gzip_mrc_name) as mrc:
            assert repr(mrc) == ("GzipMrcFile('{0}', mode='r')"
                                 .format(self.gzip_mrc_name))
    
    def test_bzip2_opening(self):
        with mrcfile.open(self.bzip2_mrc_name) as mrc:
            assert repr(mrc) == ("Bzip2MrcFile('{0}', mode='r')"
                                 .format(self.bzip2_mrc_name))
    
    def test_mmap_opening(self):
        with mrcfile.mmap(self.example_mrc_name) as mrc:
            assert repr(mrc) == ("MrcMemmap('{0}', mode='r')"
                                 .format(self.example_mrc_name))
    
    def test_new_empty_file(self):
        with mrcfile.new(self.temp_mrc_name) as mrc:
            assert repr(mrc) == ("MrcFile('{0}', mode='w+')"
                                 .format(self.temp_mrc_name))

    def test_new_empty_file_with_open_function(self):
        with mrcfile.open(self.temp_mrc_name, mode='w+') as mrc:
            assert repr(mrc) == ("MrcFile('{0}', mode='w+')"
                                 .format(self.temp_mrc_name))

    def test_error_overwriting_file_with_open_function(self):
        # Convert name to str to avoid TypeErrors in Python 2.7 with pathlib
        temp_mrc_name_str = str(self.temp_mrc_name)
        assert not os.path.exists(temp_mrc_name_str)
        open(temp_mrc_name_str, 'w+').close()
        assert os.path.exists(temp_mrc_name_str)
        with self.assertRaisesRegex(ValueError, r"call 'mrcfile\.new\(\)'"):
            mrcfile.open(self.temp_mrc_name, mode='w+')

    def test_header_only_opening(self):
        with mrcfile.open(self.example_mrc_name, header_only=True) as mrc:
            assert mrc.header is not None
            assert mrc.extended_header is not None
            assert mrc.data is None
    
    def test_opening_nonexistent_file(self):
        with self.assertRaisesRegex(Exception, "No such file"):
            mrcfile.open('no_file')
    
    def test_new_file_with_data(self):
        data = np.arange(24, dtype=np.uint16).reshape(2, 3, 4)
        with mrcfile.new(self.temp_mrc_name, data) as mrc:
            np.testing.assert_array_equal(data, mrc.data)
    
    def test_new_gzip_file(self):
        data = np.arange(24, dtype=np.uint16).reshape(4, 3, 2)
        with mrcfile.new(self.temp_mrc_name, data, compression='gzip') as mrc:
            np.testing.assert_array_equal(data, mrc.data)
            assert repr(mrc) == ("GzipMrcFile('{0}', mode='w+')"
                                 .format(self.temp_mrc_name))
    
    def test_new_bzip2_file(self):
        data = np.arange(24, dtype=np.uint16).reshape(4, 3, 2)
        with mrcfile.new(self.temp_mrc_name, data, compression='bzip2') as mrc:
            np.testing.assert_array_equal(data, mrc.data)
            assert repr(mrc) == ("Bzip2MrcFile('{0}', mode='w+')"
                                 .format(self.temp_mrc_name))
    
    def test_unknown_compression_type(self):
        with self.assertRaisesRegex(ValueError, 'Unknown compression format'):
            mrcfile.new(self.temp_mrc_name, compression='other')
    
    def test_overwriting_flag(self):
        assert not os.path.exists(str(self.temp_mrc_name))
        open(str(self.temp_mrc_name), 'w+').close()
        assert os.path.exists(str(self.temp_mrc_name))
        with self.assertRaisesRegex(ValueError, "already exists"):
            mrcfile.new(self.temp_mrc_name)
        with self.assertRaisesRegex(ValueError, "already exists"):
            mrcfile.new(self.temp_mrc_name, overwrite=False)
        mrcfile.new(self.temp_mrc_name, overwrite=True).close()
    
    def test_invalid_mode_raises_exception(self):
        with self.assertRaisesRegex(ValueError, "Mode 'z' not supported"):
            mrcfile.open(self.example_mrc_name, mode='z')
    
    def test_non_mrc_file_raises_exception(self):
        name = os.path.join(self.test_data, 'emd_3197.png')
        with self.assertRaisesRegex(ValueError, 'Map ID string not found'):
            mrcfile.open(name)
    
    def test_gzipped_non_mrc_file_raises_exception(self):
        name = os.path.join(self.test_data, 'emd_3197.png.gz')
        with self.assertRaisesRegex(ValueError, 'Map ID string not found'):
            mrcfile.open(name)
    
    def test_error_in_gzip_opening_raises_new_exception(self):
        # Tricky to test this case. Easiest to monkey-patch GzipMrcFile.__init__
        old_init = GzipMrcFile.__init__
        try:
            msg = 'Fake error: valid gzip file with invalid MRC data'
            def error(*args, **kwargs):
                raise IOError(msg)
            GzipMrcFile.__init__ = error
            with self.assertRaisesRegex(IOError, msg):
                mrcfile.open(self.gzip_mrc_name)
        finally:
            GzipMrcFile.__init__ = old_init
    
    def test_error_in_bzip2_opening_raises_new_exception(self):
        # Tricky to test this case. Easiest to monkey-patch Bzip2MrcFile.__init__
        old_init = Bzip2MrcFile.__init__
        try:
            msg = 'Fake error: valid bzip2 file with invalid MRC data'
            def error(*args, **kwargs):
                raise IOError(msg)
            Bzip2MrcFile.__init__ = error
            with self.assertRaisesRegex(IOError, msg):
                mrcfile.open(self.bzip2_mrc_name)
        finally:
            Bzip2MrcFile.__init__ = old_init
    
    def test_switching_mode(self):
        with mrcfile.new(self.temp_mrc_name) as mrc:
            mrc.set_data(np.arange(12, dtype=np.int8).reshape(3, 4))
        with mrcfile.open(self.temp_mrc_name, mode='r+') as mrc:
            mrc.set_data(np.arange(20, dtype=np.int16).reshape(2, 2, 5))
            assert mrc.header.mode == 1

    def test_simple_async_opening(self):
        with mrcfile.open_async(self.example_mrc_name).result() as mrc:
            assert repr(mrc) == ("MrcFile('{0}', mode='r')"
                                 .format(self.example_mrc_name))

    def test_slow_async_opening(self):
        # This test relies on the fact that file opening takes longer than the
        # assertions tested immediately after the open_async() call. It seems
        # to work well even with the small gzip or bzip2 files, but we use a
        # much larger gzip file to make sure.
        future = mrcfile.open_async(self.slow_mrc_name)
        assert future.running()
        assert not future.done()
        with future.result() as mrc:
            assert future.done()
            assert not future.running()
            assert repr(mrc) == ("GzipMrcFile('{0}', mode='r')"
                                 .format(self.slow_mrc_name))
        assert future.exception() is None

    def test_new_mmap(self):
        with mrcfile.new_mmap(self.temp_mrc_name,
                              (3, 4, 5, 6),
                              mrc_mode=2,
                              fill=1.1) as mrc:
            assert repr(mrc) == ("MrcMemmap('{0}', mode='w+')"
                                 .format(self.temp_mrc_name))
            assert mrc.data.shape == (3, 4, 5, 6)
            assert np.all(mrc.data == 1.1)
            assert mrc.header.nx == 6
            mrc._iostream.seek(0, os.SEEK_END)
            file_size = mrc._iostream.tell()
            assert file_size == mrc.header.nbytes + mrc.data.nbytes

    def test_new_mmap_with_extended_header(self):
        with mrcfile.new_mmap(self.temp_mrc_name,
                              (3, 4, 5),
                              mrc_mode=2,
                              fill=1.1,
                              extended_header=np.zeros(1, dtype=np.int8),
                              exttyp='TYPE') as mrc:
            assert repr(mrc) == ("MrcMemmap('{0}', mode='w+')"
                                 .format(self.temp_mrc_name))
            assert mrc.data.shape == (3, 4, 5)
            assert np.all(mrc.data == 1.1)
            assert mrc.header.nx == 5
            assert mrc.header.nsymbt == 1
            assert mrc.extended_header.nbytes == 1
            assert mrc.header.exttyp == b'TYPE'
            mrc._iostream.seek(0, os.SEEK_END)
            file_size = mrc._iostream.tell()
            exp_size = mrc.header.nbytes + mrc.extended_header.nbytes + mrc.data.nbytes
            assert file_size == exp_size

    def test_new_mmap_with_aligned_offset(self):
        # Test for awkward situation in https://github.com/ccpem/mrcfile/issues/65
        # A zero-size mmap is created during the process of creating a new MrcMemmap
        # object. If the header size is an exact multiple of mmap.ALLOCATIONGRANULARITY,
        # mmap creation fails with an error.
        problem_ext_head_size = mmap.ALLOCATIONGRANULARITY - 1024
        with mrcfile.new_mmap(self.temp_mrc_name,
                              (3, 4, 5),
                              mrc_mode=2,
                              fill=1.1,
                              extended_header=np.zeros(problem_ext_head_size,
                                                       dtype=np.int8)) as mrc:
            assert repr(mrc) == ("MrcMemmap('{0}', mode='w+')"
                                 .format(self.temp_mrc_name))
            assert mrc.data.shape == (3, 4, 5)
            assert np.all(mrc.data == 1.1)
            assert mrc.header.nx == 5
            assert mrc.header.nsymbt == problem_ext_head_size
            assert mrc.extended_header.nbytes == problem_ext_head_size

            # The workaround for this bug adds one extra byte to the file to avoid
            # triggering the mmap error. After a data array has been set, the file size
            # should be normal again.
            mrc._iostream.seek(0, os.SEEK_END)
            file_size = mrc._iostream.tell()
            exp_size = mrc.header.nbytes + mrc.extended_header.nbytes + mrc.data.nbytes
            assert file_size == exp_size

    @unittest.skipIf(sys.maxsize <= np.iinfo(np.int32).max, "can't run test on 32-bit")
    def test_new_mmap_with_extended_header_too_big(self):
        ext = np.empty((int(np.iinfo(np.int32).max) + 1,), dtype='V1')
        with self.assertRaisesRegex(ValueError, "extended header is too large"):
            mrcfile.new_mmap(self.temp_mrc_name,
                             (3, 4, 5, 6),
                             mrc_mode=2,
                             fill=1.1,
                             extended_header=ext)

    def test_new_mmap_with_shape_too_big(self):
        with self.assertRaisesRegex(ValueError, "shape is too large"):
            mrcfile.new_mmap(self.temp_mrc_name,
                             (np.iinfo(np.int32).max + 1, 1))

    def test_write(self):
        data_in = np.random.random((10, 10)).astype(np.float16)
        mrcfile.write(self.temp_mrc_name, data_in, voxel_size=1.1)
        with mrcfile.open(self.temp_mrc_name) as mrc:
            data_out = mrc.data
            voxel_size = mrc.voxel_size
        assert np.allclose(data_in, data_out)
        self.assertAlmostEqual(voxel_size.x, 1.1)
        self.assertAlmostEqual(voxel_size.y, 1.1)
        self.assertAlmostEqual(voxel_size.z, 1.1)

    def test_write_with_auto_compression(self):
        data_in = np.random.random((10, 10)).astype(np.float16)
        mrcfile.write(self.temp_gz_mrc_name, data_in)
        with mrcfile.open(self.temp_gz_mrc_name) as mrc:
            assert isinstance(mrc, GzipMrcFile)


@unittest.skipIf(pathlib_unavailable, "pathlib not available")
class LoadFunctionTestWithPathlib(LoadFunctionTest):

    """Class to run the load function tests using pathlib paths instead of strings."""

    def setUp(self):
        super(LoadFunctionTestWithPathlib, self).setUp()
        self.temp_mrc_name = Path(self.temp_mrc_name)
        self.temp_gz_mrc_name = Path(self.temp_gz_mrc_name)
        self.example_mrc_name = Path(self.example_mrc_name)
        self.gzip_mrc_name = Path(self.gzip_mrc_name)
        self.bzip2_mrc_name = Path(self.bzip2_mrc_name)
        self.slow_mrc_name = Path(self.slow_mrc_name)


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2016, Science and Technology Facilities Council
# This software is distributed under a BSD licence. See LICENSE.txt.

"""
Tests for mrcmemmap.py
"""

# Import Python 3 features for future-proofing
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import os
import unittest

import numpy as np

from . import test_mrcfile
from mrcfile.mrcmemmap import MrcMemmap


class MrcMemmapTest(test_mrcfile.MrcFileTest):
    
    """Unit tests for MRC file I/O with memory-mapped files.
    
    Note that this test class inherits MrcFileTest to ensure all of the tests
    for MrcObject and MrcFile work correctly for the MrcMemmap subclass.
    
    """
    
    def setUp(self):
        # Set up as if for MrcFileTest
        super(MrcMemmapTest, self).setUp()
        
        # Set the newmrc method to the MrcMemmap constructor
        self.newmrc = MrcMemmap
        
        # Set up parameters so MrcObject tests run on the MrcMemmap class
        obj_mrc_name = os.path.join(self.test_output, 'test_mrcobject.mrc')
        self.mrcobject = MrcMemmap(obj_mrc_name, 'w+', overwrite=True)
    
    def test_repr(self):
        """Override test to change expected repr string."""
        with MrcMemmap(self.example_mrc_name) as mrc:
            assert repr(mrc) == "MrcMemmap('{0}', mode='r')".format(self.example_mrc_name)
    
    def test_exception_raised_if_file_is_too_small_for_reading_data(self):
        """Override test to change expected error message."""
        with self.newmrc(self.temp_mrc_name, mode='w+') as mrc:
            mrc.set_data(np.arange(24, dtype=np.int16).reshape(2, 3, 4))
            assert mrc.header.mz == 2
            mrc.header.mz = mrc.header.nz = 3
        # The exception type and message are different on Linux and Windows
        expected_error_msg = ("mmap length is greater than file size"
                              "|Not enough storage is available"
                              "|Not enough memory resources are available")
        with self.assertRaisesRegex(Exception, expected_error_msg):
            self.newmrc(self.temp_mrc_name)
    
    def test_data_is_not_copied_unnecessarily(self):
        """Override test because data has to be copied for mmap."""
        data = np.arange(6, dtype=np.int16).reshape(1, 2, 3)
        self.mrcobject.set_data(data)
        assert self.mrcobject.data is not data
    
    def test_data_array_cannot_be_changed_after_closing_file(self):
        mrc = self.newmrc(self.temp_mrc_name, mode='w+')
        mrc.set_data(np.arange(12, dtype=np.int16).reshape(3, 4))
        data_ref = mrc.data
        # Check that writing to the data array does not raise an exception
        data_ref[0,0] = 1
        mrc.close()
        assert not data_ref.flags.writeable
        with self.assertRaises(ValueError):
            data_ref[0,0] = 2


if __name__ == "__main__":
    unittest.main()



# Copyright (c) 2016, Science and Technology Facilities Council
# This software is distributed under a BSD licence. See LICENSE.txt.

"""
Tests for mrcfile validation functions.
"""

# Import Python 3 features for future-proofing
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import os
import shutil
import sys
import tempfile
import unittest
import warnings

# Avoid str/unicode issues when traceback.print_exc tries to write to StringIO
# (see https://stackoverflow.com/a/34872005)
try:
    # Python 2
    from cStringIO import StringIO
except ImportError:
    # Python 3
    from io import StringIO

import numpy as np

import mrcfile
from mrcfile.validator import validate_all
from . import helpers


class ValidationTest(helpers.AssertRaisesRegexMixin, unittest.TestCase):
    
    """Unit tests for MRC validation functions.
    
    """
    
    def setUp(self):
        super(ValidationTest, self).setUp()
        
        # Set up test files and names to be used
        self.test_data = helpers.get_test_data_path()
        self.test_output = tempfile.mkdtemp()
        self.temp_mrc_name = os.path.join(self.test_output, 'test_mrcfile.mrc')
        self.example_mrc_name = os.path.join(self.test_data, 'EMD-3197.map')
        self.gzip_mrc_name = os.path.join(self.test_data, 'emd_3197.map.gz')
        self.bzip2_mrc_name = os.path.join(self.test_data, 'EMD-3197.map.bz2')
        self.ext_header_mrc_name = os.path.join(self.test_data, 'EMD-3001.map')
        self.fei1_ext_header_mrc_name = os.path.join(self.test_data, 'fei-extended.mrc')
        self.fei2_ext_header_mrc_name = os.path.join(self.test_data, 'epu2.9_example.mrc')
        self.not_an_mrc_name = os.path.join(self.test_data, 'README.txt')
        
        # Set up stream to catch print output from validate()
        self.print_stream = StringIO()
        
        # Replace stdout and stderr to capture output for checking
        self.orig_stdout = sys.stdout
        self.orig_stderr = sys.stderr
        sys.stdout = StringIO()
        sys.stderr = StringIO()
    
    def tearDown(self):
        # Restore stdout and stderr
        sys.stdout = self.orig_stdout
        sys.stderr = self.orig_stderr
        
        self.print_stream.close()
        if os.path.exists(self.test_output):
            shutil.rmtree(self.test_output)
        super(ValidationTest, self).tearDown()
    
    def test_good_file(self):
        with mrcfile.new(self.temp_mrc_name) as mrc:
            mrc.set_data(np.arange(36, dtype=np.float32).reshape(3, 3, 4))
            mrc.voxel_size = 2.3
        result = mrcfile.validate(self.temp_mrc_name, self.print_stream)
        assert result == True
        print_output = self.print_stream.getvalue()
        assert print_output == (
            "Checking if " + self.temp_mrc_name + " is a valid MRC2014 file...\n"
            "File appears to be valid.\n"
        )
        assert len(sys.stdout.getvalue()) == 0
        assert len(sys.stderr.getvalue()) == 0
    
    def test_emdb_file(self):
        result = mrcfile.validate(self.example_mrc_name, self.print_stream)
        assert result == False
        print_output = self.print_stream.getvalue()
        assert print_output.strip() == (
            "Checking if " + self.example_mrc_name + " is a valid MRC2014 file...\n"
            "File does not declare MRC format version 20140 or 20141: nversion = 0"
        )
        assert len(sys.stdout.getvalue()) == 0
        assert len(sys.stderr.getvalue()) == 0
    
    def test_gzip_emdb_file(self):
        result = mrcfile.validate(self.gzip_mrc_name, self.print_stream)
        assert result == False
        print_output = self.print_stream.getvalue()
        assert print_output.strip() == (
            "Checking if " + self.gzip_mrc_name + " is a valid MRC2014 file...\n"
            "File does not declare MRC format version 20140 or 20141: nversion = 0"
        )
        assert len(sys.stdout.getvalue()) == 0
        assert len(sys.stderr.getvalue()) == 0
    
    def test_bzip2_emdb_file(self):
        result = mrcfile.validate(self.bzip2_mrc_name, self.print_stream)
        assert result == False
        print_output = self.print_stream.getvalue()
        assert print_output.strip() == (
            "Checking if " + self.bzip2_mrc_name + " is a valid MRC2014 file...\n"
            "File does not declare MRC format version 20140 or 20141: nversion = 0"
        )
        assert len(sys.stdout.getvalue()) == 0
        assert len(sys.stderr.getvalue()) == 0
    
    def test_emdb_cryst_file(self):
        result = mrcfile.validate(self.ext_header_mrc_name, self.print_stream)
        assert result == False
        print_output = self.print_stream.getvalue()
        assert print_output.strip() == (
            "Checking if " + self.ext_header_mrc_name + " is a valid MRC2014 file...\n"
            "File does not declare MRC format version 20140 or 20141: nversion = 0\n"
            "Extended header type is undefined or unrecognised: exttyp = ''"
        )
        assert len(sys.stdout.getvalue()) == 0
        assert len(sys.stderr.getvalue()) == 0
    
    def check_temp_mrc_invalid_with_warning(self, message):
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            result = mrcfile.validate(self.temp_mrc_name,
                                      print_file=self.print_stream)
            assert result == False
            print_output = self.print_stream.getvalue()
            assert message.lower() in print_output.lower()
            assert len(w) == 1
            assert issubclass(w[0].category, RuntimeWarning)
            assert message in str(w[0].message)

    def test_incorrect_map_id(self):
        with mrcfile.new(self.temp_mrc_name) as mrc:
            mrc.header.map = b'fake'
        self.check_temp_mrc_invalid_with_warning("Map ID string")
        assert len(sys.stdout.getvalue()) == 0
        assert len(sys.stderr.getvalue()) == 0
    
    def test_short_map_id(self):
        """This tests the case of files where the map ID is almost correct.
        For example, MotionCor2 writes files with ID 'MAP\0', which is not
        valid according to the MRC2014 spec on the CCP-EM website, but could
        be considered valid according to the MRC2014 paper (which just
        specifies 'MAP', i.e. without the final byte). We should read such
        files without errors or warnings, but they should fail a strict
        validation check."""
        with mrcfile.new(self.temp_mrc_name) as mrc:
            mrc.header.map = b'MAP\0'
        result = mrcfile.validate(self.temp_mrc_name,
                                  print_file=self.print_stream)
        assert result == False
        print_output = self.print_stream.getvalue()
        assert "Map ID string is incorrect" in print_output
        assert len(sys.stdout.getvalue()) == 0
        assert len(sys.stderr.getvalue()) == 0
    
    def test_incorrect_machine_stamp(self):
        with mrcfile.new(self.temp_mrc_name) as mrc:
            mrc.header.machst = bytearray(b'    ')
        self.check_temp_mrc_invalid_with_warning("machine stamp")
        assert len(sys.stdout.getvalue()) == 0
        assert len(sys.stderr.getvalue()) == 0
    
    def test_invalid_mode(self):
        with mrcfile.new(self.temp_mrc_name) as mrc:
            mrc.set_data(np.arange(12, dtype=np.float32).reshape(1, 3, 4))
            mrc.header.mode = 8
        self.check_temp_mrc_invalid_with_warning("mode")
        assert len(sys.stdout.getvalue()) == 0
        assert len(sys.stderr.getvalue()) == 0
    
    def test_file_too_small(self):
        with mrcfile.new(self.temp_mrc_name) as mrc:
            mrc.set_data(np.arange(12, dtype=np.float32).reshape(1, 3, 4))
            mrc.header.nz = 2
        self.check_temp_mrc_invalid_with_warning("data block")
        assert len(sys.stdout.getvalue()) == 0
        assert len(sys.stderr.getvalue()) == 0
    
    def test_file_too_large(self):
        with mrcfile.new(self.temp_mrc_name) as mrc:
            mrc.set_data(np.arange(36, dtype=np.float32).reshape(3, 3, 4))
            mrc.header.nz = 2
        self.check_temp_mrc_invalid_with_warning("larger than expected")
        assert len(sys.stdout.getvalue()) == 0
        assert len(sys.stderr.getvalue()) == 0
    
    def test_negative_mx(self):
        with mrcfile.new(self.temp_mrc_name) as mrc:
            mrc.header.mx = -10
        result = mrcfile.validate(self.temp_mrc_name,
                                  print_file=self.print_stream)
        assert result == False
        print_output = self.print_stream.getvalue()
        assert "Header field 'mx' is negative" in print_output
        assert len(sys.stdout.getvalue()) == 0
        assert len(sys.stderr.getvalue()) == 0
    
    def test_negative_my(self):
        with mrcfile.new(self.temp_mrc_name) as mrc:
            mrc.header.my = -10
        result = mrcfile.validate(self.temp_mrc_name,
                                  print_file=self.print_stream)
        assert result == False
        print_output = self.print_stream.getvalue()
        assert "Header field 'my' is negative" in print_output
        assert len(sys.stdout.getvalue()) == 0
        assert len(sys.stderr.getvalue()) == 0
    
    def test_negative_mz(self):
        with mrcfile.new(self.temp_mrc_name) as mrc:
            mrc.header.mz = -10
        result = mrcfile.validate(self.temp_mrc_name,
                                  print_file=self.print_stream)
        assert result == False
        print_output = self.print_stream.getvalue()
        assert "Header field 'mz' is negative" in print_output
        assert len(sys.stdout.getvalue()) == 0
        assert len(sys.stderr.getvalue()) == 0
    
    def test_negative_ispg(self):
        with mrcfile.new(self.temp_mrc_name) as mrc:
            mrc.header.ispg = -10
        result = mrcfile.validate(self.temp_mrc_name,
                                  print_file=self.print_stream)
        assert result == False
        print_output = self.print_stream.getvalue()
        assert "Header field 'ispg' is negative" in print_output
        assert len(sys.stdout.getvalue()) == 0
        assert len(sys.stderr.getvalue()) == 0
    
    def test_negative_nlabl(self):
        with mrcfile.new(self.temp_mrc_name) as mrc:
            mrc.header.nlabl = -3
        result = mrcfile.validate(self.temp_mrc_name,
                                  print_file=self.print_stream)
        assert result == False
        print_output = self.print_stream.getvalue()
        assert "Header field 'nlabl' is negative" in print_output
        assert len(sys.stdout.getvalue()) == 0
        assert len(sys.stderr.getvalue()) == 0
    
    def test_negative_cella_x(self):
        with mrcfile.new(self.temp_mrc_name) as mrc:
            mrc.header.cella.x = -10
        result = mrcfile.validate(self.temp_mrc_name,
                                  print_file=self.print_stream)
        assert result == False
        print_output = self.print_stream.getvalue()
        assert "Cell dimension 'x' is negative" in print_output
        assert len(sys.stdout.getvalue()) == 0
        assert len(sys.stderr.getvalue()) == 0
    
    def test_invalid_axis_mapping(self):
        with mrcfile.new(self.temp_mrc_name) as mrc:
            mrc.header.mapc = 3
            mrc.header.mapr = 4
            mrc.header.maps = -200
        result = mrcfile.validate(self.temp_mrc_name,
                                  print_file=self.print_stream)
        assert result == False
        print_output = self.print_stream.getvalue()
        assert "Invalid axis mapping: found [-200, 3, 4]" in print_output
        assert len(sys.stdout.getvalue()) == 0
        assert len(sys.stderr.getvalue()) == 0
    
    def test_mz_correct_for_volume_stack(self):
        with mrcfile.new(self.temp_mrc_name) as mrc:
            mrc.set_data(np.arange(120, dtype=np.float32).reshape(3, 2, 4, 5))
        with warnings.catch_warnings(record=True):
            result = mrcfile.validate(self.temp_mrc_name,
                                      print_file=self.print_stream)
            assert result == True
        assert len(sys.stdout.getvalue()) == 0
        assert len(sys.stderr.getvalue()) == 0
    
    def test_mz_incorrect_for_volume_stack(self):
        with mrcfile.new(self.temp_mrc_name) as mrc:
            mrc.set_data(np.arange(120, dtype=np.float32).reshape(3, 2, 4, 5))
            mrc.header.mz = 5
        with warnings.catch_warnings(record=True):
            result = mrcfile.validate(self.temp_mrc_name,
                                      print_file=self.print_stream)
            assert result == False
            print_output = self.print_stream.getvalue()
            assert ("Error in dimensions for volume stack: nz should be "
                    "divisible by mz. Found nz = 6, mz = 5" in print_output)
        assert len(sys.stdout.getvalue()) == 0
        assert len(sys.stderr.getvalue()) == 0
    
    def test_nlabl_too_large(self):
        with mrcfile.new(self.temp_mrc_name) as mrc:
            mrc.header.label[1] = 'test label'
            mrc.header.nlabl = 3
        result = mrcfile.validate(self.temp_mrc_name,
                                  print_file=self.print_stream)
        assert result == False
        print_output = self.print_stream.getvalue()
        assert ("Error in header labels: "
                "nlabl is 3 but 2 labels contain text" in print_output)
        assert len(sys.stdout.getvalue()) == 0
        assert len(sys.stderr.getvalue()) == 0
    
    def test_nlabl_too_small(self):
        with mrcfile.new(self.temp_mrc_name) as mrc:
            mrc.header.label[1] = 'test label'
            mrc.header.nlabl = 1
        result = mrcfile.validate(self.temp_mrc_name,
                                  print_file=self.print_stream)
        assert result == False
        print_output = self.print_stream.getvalue()
        assert ("Error in header labels: "
                "nlabl is 1 but 2 labels contain text" in print_output)
        assert len(sys.stdout.getvalue()) == 0
        assert len(sys.stderr.getvalue()) == 0
    
    def test_empty_labels_in_list(self):
        with mrcfile.new(self.temp_mrc_name) as mrc:
            mrc.header.label[2] = 'test label'
            mrc.header.nlabl = 2
        result = mrcfile.validate(self.temp_mrc_name,
                                  print_file=self.print_stream)
        assert result == False
        print_output = self.print_stream.getvalue()
        assert ("Error in header labels: empty labels appear between "
                "text-containing labels" in print_output)
        assert len(sys.stdout.getvalue()) == 0
        assert len(sys.stderr.getvalue()) == 0

    def test_old_format_version(self):
        with mrcfile.new(self.temp_mrc_name) as mrc:
            mrc.header.nversion = 20140
        result = mrcfile.validate(self.temp_mrc_name,
                                  print_file=self.print_stream)
        assert result == True
        print_output = self.print_stream.getvalue()
        assert len(sys.stdout.getvalue()) == 0
        assert len(sys.stderr.getvalue()) == 0
    
    def test_incorrect_format_version(self):
        with mrcfile.new(self.temp_mrc_name) as mrc:
            mrc.header.nversion = 20139
        result = mrcfile.validate(self.temp_mrc_name,
                                  print_file=self.print_stream)
        assert result == False
        print_output = self.print_stream.getvalue()
        assert "File does not declare MRC format version 20140 or 20141" in print_output
        assert len(sys.stdout.getvalue()) == 0
        assert len(sys.stderr.getvalue()) == 0
    
    def test_missing_exttyp(self):
        with mrcfile.new(self.temp_mrc_name) as mrc:
            mrc.set_extended_header(np.arange(10))
        result = mrcfile.validate(self.temp_mrc_name,
                                  print_file=self.print_stream)
        assert result == False
        print_output = self.print_stream.getvalue()
        assert "Extended header type is undefined or unrecognised" in print_output
        assert len(sys.stdout.getvalue()) == 0
        assert len(sys.stderr.getvalue()) == 0
    
    def test_unknown_exttyp(self):
        with mrcfile.new(self.temp_mrc_name) as mrc:
            mrc.set_extended_header(np.arange(10))
            mrc.header.exttyp = 'Fake'
        result = mrcfile.validate(self.temp_mrc_name,
                                  print_file=self.print_stream)
        assert result == False
        print_output = self.print_stream.getvalue()
        assert "Extended header type is undefined or unrecognised" in print_output
        assert len(sys.stdout.getvalue()) == 0
        assert len(sys.stderr.getvalue()) == 0
    
    def test_incorrect_rms(self):
        data = np.arange(-10, 20, dtype=np.float32).reshape(2, 3, 5)
        with mrcfile.new(self.temp_mrc_name) as mrc:
            mrc.set_data(data)
            mrc.header.rms = 9.0
        result = mrcfile.validate(self.temp_mrc_name,
                                  print_file=self.print_stream)
        assert result == False
        print_output = self.print_stream.getvalue()
        assert ("Data statistics appear to be inaccurate: RMS deviation is {0} but the"
                " value in the header is 9.0".format(data.std()) in print_output)
        assert len(sys.stdout.getvalue()) == 0
        assert len(sys.stderr.getvalue()) == 0
    
    def test_rms_undetermined(self):
        data = np.arange(-10, 20, dtype=np.float32).reshape(2, 3, 5)
        with mrcfile.new(self.temp_mrc_name) as mrc:
            mrc.set_data(data)
            mrc.header.rms = -15
        result = mrcfile.validate(self.temp_mrc_name,
                                  print_file=self.print_stream)
        assert result == True
        assert len(sys.stdout.getvalue()) == 0
        assert len(sys.stderr.getvalue()) == 0
    
    def test_incorrect_dmin(self):
        data = np.arange(-10, 20, dtype=np.float32).reshape(2, 3, 5)
        with mrcfile.new(self.temp_mrc_name) as mrc:
            mrc.set_data(data)
            mrc.header.dmin = -11
        result = mrcfile.validate(self.temp_mrc_name,
                                  print_file=self.print_stream)
        assert result == False
        print_output = self.print_stream.getvalue()
        assert ("Data statistics appear to be inaccurate: minimum is {0} but the value"
                " in the header is -11".format(data.min()) in print_output)
        assert len(sys.stdout.getvalue()) == 0
        assert len(sys.stderr.getvalue()) == 0
    
    def test_incorrect_dmax(self):
        data = np.arange(-10, 20, dtype=np.float32).reshape(2, 3, 5)
        with mrcfile.new(self.temp_mrc_name) as mrc:
            mrc.set_data(data)
            mrc.header.dmax = 15
        result = mrcfile.validate(self.temp_mrc_name,
                                  print_file=self.print_stream)
        assert result == False
        print_output = self.print_stream.getvalue()
        assert ("Data statistics appear to be inaccurate: maximum is {0} but the value"
                " in the header is 15".format(data.max()) in print_output)
        assert len(sys.stdout.getvalue()) == 0
        assert len(sys.stderr.getvalue()) == 0
    
    def test_min_and_max_undetermined(self):
        data = np.arange(-10, 20, dtype=np.float32).reshape(2, 3, 5)
        with mrcfile.new(self.temp_mrc_name) as mrc:
            mrc.set_data(data)
            mrc.header.dmin = 30.1
            mrc.header.dmax = 30.0
        result = mrcfile.validate(self.temp_mrc_name,
                                  print_file=self.print_stream)
        assert result == True
        assert len(sys.stdout.getvalue()) == 0
        assert len(sys.stderr.getvalue()) == 0
    
    def test_incorrect_dmean(self):
        data = np.arange(-10, 20, dtype=np.float32).reshape(2, 3, 5)
        with mrcfile.new(self.temp_mrc_name) as mrc:
            mrc.set_data(data)
            mrc.header.dmean = -2.5
        result = mrcfile.validate(self.temp_mrc_name,
                                  print_file=self.print_stream)
        assert result == False
        print_output = self.print_stream.getvalue()
        assert ("Data statistics appear to be inaccurate: mean is {0} but the value in"
                " the header is -2.5".format(data.mean(dtype=np.float64))
                in print_output)
        assert len(sys.stdout.getvalue()) == 0
        assert len(sys.stderr.getvalue()) == 0
    
    def test_incorrect_dmean_with_undetermined_dmin_and_dmax(self):
        data = np.arange(-10, 20, dtype=np.float32).reshape(2, 3, 5)
        with mrcfile.new(self.temp_mrc_name) as mrc:
            mrc.set_data(data)
            mrc.header.dmin = 20
            mrc.header.dmax = -30.1
            mrc.header.dmean = -2.5
        result = mrcfile.validate(self.temp_mrc_name,
                                  print_file=self.print_stream)
        assert result == False
        print_output = self.print_stream.getvalue()
        assert ("Data statistics appear to be inaccurate: mean is {0} but the value in"
                " the header is -2.5".format(data.mean(dtype=np.float64))
                in print_output)
        assert len(sys.stdout.getvalue()) == 0
        assert len(sys.stderr.getvalue()) == 0
    
    def test_mean_undetermined(self):
        data = np.arange(-10, 20, dtype=np.float32).reshape(2, 3, 5)
        with mrcfile.new(self.temp_mrc_name) as mrc:
            mrc.set_data(data)
            mrc.header.dmean = -11
        result = mrcfile.validate(self.temp_mrc_name,
                                  print_file=self.print_stream)
        assert result == True
        assert len(sys.stdout.getvalue()) == 0
        assert len(sys.stderr.getvalue()) == 0
    
    def test_min_max_and_mean_undetermined(self):
        data = np.arange(-10, 20, dtype=np.float32).reshape(2, 3, 5)
        with mrcfile.new(self.temp_mrc_name) as mrc:
            mrc.set_data(data)
            mrc.header.dmin = 30.1
            mrc.header.dmax = 30.0
            mrc.header.dmean = 29.9
        result = mrcfile.validate(self.temp_mrc_name,
                                  print_file=self.print_stream)
        assert result == True
        assert len(sys.stdout.getvalue()) == 0
        assert len(sys.stderr.getvalue()) == 0
    
    def test_many_problems_simultaneously(self):
        data = np.arange(-10, 20, dtype=np.float32).reshape(3, 2, 5)
        with mrcfile.new(self.temp_mrc_name) as mrc:
            mrc.set_data(data)
            mrc.set_extended_header(data)
            mrc.header.nz = 2
            mrc.header.my = -1000
            mrc.header.mz = -5
            mrc.header.cella.y = -12.1
            mrc.header.mapc = 5
            mrc.header.dmin = 10
            mrc.header.dmax = 11
            mrc.header.dmean = 19.0
            mrc.header.ispg = -20
            mrc.header.exttyp = 'fake'
            mrc.header.nversion = 0
            mrc.header.rms = 0.0
            mrc.header.nlabl = 4
            mrc.header.label[9] = 'test label'
        with warnings.catch_warnings(record=True):
            result = mrcfile.validate(self.temp_mrc_name,
                                      print_file=self.print_stream)
            assert result == False
            print_output = self.print_stream.getvalue()
            assert len(print_output.split('\n')) == 16
        assert len(sys.stdout.getvalue()) == 0
        assert len(sys.stderr.getvalue()) == 0
    
    def create_good_files(self):
        """Create some files known to be valid and return their names."""
        good_mrc_name_1 = os.path.join(self.test_output, 'good_file_1.mrc')
        good_mrc_name_2 = os.path.join(self.test_output, 'good_file_2.mrc')
        
        # Make good files which will pass validation
        with mrcfile.new(good_mrc_name_1) as mrc:
            mrc.set_data(np.arange(36, dtype=np.float32).reshape(3, 3, 4))
        
        with mrcfile.new(good_mrc_name_2) as mrc:
            mrc.set_data(np.arange(36, dtype=np.uint16).reshape(3, 3, 4))
        
        return [
            good_mrc_name_1,
            good_mrc_name_2,
            self.fei1_ext_header_mrc_name,
            self.fei2_ext_header_mrc_name
        ]
    
    def test_validate_good_files(self):
        good_files = self.create_good_files()
        result = validate_all(good_files, print_file=self.print_stream)
        assert result == True
        print_output = self.print_stream.getvalue()
        assert print_output == (
            "Checking if " + good_files[0] + " is a valid MRC2014 file...\n"
            "File appears to be valid.\n"
            "Checking if " + good_files[1] + " is a valid MRC2014 file...\n"
            "File appears to be valid.\n"
            "Checking if " + good_files[2] + " is a valid MRC2014 file...\n"
            "File appears to be valid.\n"
            "Checking if " + good_files[3] + " is a valid MRC2014 file...\n"
            "File appears to be valid.\n"
        )
        assert len(sys.stdout.getvalue()) == 0
        assert len(sys.stderr.getvalue()) == 0
    
    def test_validate_bad_files(self):
        bad_files = [
            self.not_an_mrc_name,
            self.example_mrc_name,
            self.ext_header_mrc_name,
            self.gzip_mrc_name
        ]
        result = validate_all(bad_files, print_file=self.print_stream)
        assert result == False
        print_output = self.print_stream.getvalue()
        assert len(print_output) > 0
        assert "Checking if " + bad_files[0] + " is a valid MRC2014 file..." in print_output
        assert "Checking if " + bad_files[1] + " is a valid MRC2014 file..." in print_output
        assert "Checking if " + bad_files[2] + " is a valid MRC2014 file..." in print_output
        assert "Checking if " + bad_files[3] + " is a valid MRC2014 file..." in print_output
        assert "ValueError:" in print_output
        assert len(sys.stdout.getvalue()) == 0
        assert len(sys.stderr.getvalue()) == 0
     
    def test_validate_good_and_bad_files(self):
        files = self.create_good_files() + [
            self.not_an_mrc_name,
            self.example_mrc_name,
            self.ext_header_mrc_name,
            self.gzip_mrc_name
        ]
        result = validate_all(files, print_file=self.print_stream)
        assert result == False
        print_output = self.print_stream.getvalue()
        assert len(print_output) > 0
        assert len(sys.stdout.getvalue()) == 0
        assert len(sys.stderr.getvalue()) == 0


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2016, Science and Technology Facilities Council
# This software is distributed under a BSD licence. See LICENSE.txt.

"""
Tests for mrcfile.py
"""

# Import Python 3 features for future-proofing
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import os
import shutil
import sys
import tempfile
import unittest
import warnings

import numpy as np
try:
    from numpy.exceptions import ComplexWarning
except ImportError:
    from numpy import ComplexWarning

from . import helpers, test_mrcobject
from mrcfile.mrcfile import MrcFile
from mrcfile.mrcobject import (IMAGE_STACK_SPACEGROUP, VOLUME_SPACEGROUP,
                               VOLUME_STACK_SPACEGROUP)
import mrcfile.utils as utils

# Try to import pathlib if we can
pathlib_unavailable = False
try:
    from pathlib import Path
except ImportError:
    try:
        from pathlib2 import Path
    except ImportError:
        pathlib_unavailable = True


# Doctest stuff commented out for now - would be nice to get it working!
# import doctest

# doc_test_dir = tempfile.mkdtemp()
# doc_test_file = MrcFile(os.path.join(doc_test_dir, 'doc_test.mrc'), 'w+')
#
# def tearDownModule():
#     global doc_test_dir, doc_test_file
#     doc_test_file.close()
#     if os.path.exists(doc_test_dir):
#         shutil.rmtree(doc_test_dir)
#
# def load_tests(loader, tests, ignore):
#     global doc_test_file
#     tests.addTests(doctest.DocTestSuite(mrcfile, extraglobs={'mrc': doc_test_file}))
#     return tests


class MrcFileTest(test_mrcobject.MrcObjectTest):

    """Unit tests for MRC file I/O.

    Note that this test class inherits MrcObjectTest to ensure all of the tests
    for MrcObject work correctly for the MrcFile subclass. setUp() is a little
    more complicated as a result.

    """

    def setUp(self):
        super(MrcFileTest, self).setUp()

        # Set up test files and names to be used
        self.test_data = helpers.get_test_data_path()
        self.test_output = tempfile.mkdtemp()
        self.temp_mrc_name = os.path.join(self.test_output, 'test_mrcfile.mrc')
        self.example_mrc_name = os.path.join(self.test_data, 'EMD-3197.map')
        self.ext_header_mrc_name = os.path.join(self.test_data, 'EMD-3001.map')
        self.fei1_ext_header_mrc_name = os.path.join(self.test_data, 'fei-extended.mrc')
        self.fei2_ext_header_mrc_name = os.path.join(self.test_data, 'epu2.9_example.mrc')

        # Set newmrc method as MrcFile constructor, to allow override by subclasses
        self.newmrc = MrcFile

        # Set up parameters so MrcObject tests run on the MrcFile class
        obj_mrc_name = os.path.join(self.test_output, 'test_mrcobject.mrc')
        self.mrcobject = MrcFile(obj_mrc_name, 'w+')

    def tearDown(self):
        self.mrcobject.close()
        if os.path.exists(self.test_output):
            shutil.rmtree(self.test_output)
        super(MrcFileTest, self).tearDown()

    ############################################################################
    #
    # Tests which depend on existing files (in the test_data directory)
    #

    def test_machine_stamp_is_read_correctly(self):
        with self.newmrc(self.example_mrc_name) as mrc:
            assert np.array_equal(mrc.header.machst, [ 0x44, 0x41, 0, 0 ])
            if sys.byteorder == 'little':
                assert mrc.header.mode.dtype.byteorder in ('=', '<')
                assert mrc.data.dtype.byteorder in ('=', '<')
            else:
                assert mrc.header.mode.dtype.byteorder == '<'
                assert mrc.data.dtype.byteorder == '<'

    def test_non_mrc_file_is_rejected(self):
        name = os.path.join(self.test_data, 'emd_3197.png')
        with (self.assertRaisesRegex(ValueError, 'Map ID string not found')):
            self.newmrc(name)

    def test_non_mrc_file_gives_correct_warnings_in_permissive_mode(self):
        name = os.path.join(self.test_data, 'emd_3197.png')
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            with self.newmrc(name, permissive=True) as mrc:
                assert mrc.header is not None
                assert mrc.extended_header is None
                assert mrc.data is None
            assert len(w) == 4
            assert "Map ID string not found" in str(w[0].message)
            assert "Unrecognised machine stamp" in str(w[1].message)
            assert "Expected 976237114 bytes in extended header" in str(w[2].message)
            assert "Unrecognised mode" in str(w[3].message)

    def test_repr(self):
        with self.newmrc(self.example_mrc_name) as mrc:
            expected = "MrcFile('{0}', mode='r')".format(self.example_mrc_name)
            assert repr(mrc) == expected

    def test_data_values_are_correct(self):
        with self.newmrc(self.example_mrc_name) as mrc:
            # Check a few values
            self.assertAlmostEqual(mrc.data[0, 0, 0], -1.8013091)
            self.assertAlmostEqual(mrc.data[9, 6, 13], 4.6207790)
            self.assertAlmostEqual(mrc.data[9, 6, 14], 5.0373931)
            self.assertAlmostEqual(mrc.data[-1, -1, -1], 1.3078574)

            # Calculate some statistics for all values
            calc_max = mrc.data.max()
            calc_min = mrc.data.min()
            calc_mean = mrc.data.mean(dtype=np.float64)
            calc_std = mrc.data.std()
            calc_sum = mrc.data.sum()

            # Compare calculated values with header records
            self.assertAlmostEqual(calc_max, mrc.header.dmax)
            self.assertAlmostEqual(calc_min, mrc.header.dmin)
            self.assertAlmostEqual(calc_mean, mrc.header.dmean)
            self.assertAlmostEqual(calc_std, mrc.header.rms)

            # Convert calc_sum to float to fix a bug with memmap comparisons in python 3
            self.assertAlmostEqual(float(calc_sum), 6268.896, places=3)

    def test_absent_extended_header_is_read_as_zero_length_array(self):
        with self.newmrc(self.example_mrc_name) as mrc:
            assert mrc.header.nbytes == 1024
            assert mrc.header.nsymbt == 0
            assert mrc.extended_header.nbytes == 0
            assert mrc.extended_header.dtype.kind == 'V'
            assert mrc.extended_header.tobytes() == b''

    def test_extended_header_is_read_correctly(self):
        with self.newmrc(self.ext_header_mrc_name) as mrc:
            assert mrc.header.nbytes == 1024
            assert mrc.header.nsymbt == 160
            assert mrc.extended_header.nbytes == 160
            assert mrc.extended_header.dtype.kind == 'V'
            mrc.extended_header.dtype = 'S80'
            ext = mrc.extended_header
            assert ext[0] == (b'X,  Y,  Z                               '
                              b'                                        ')
            assert ext[1] == (b'-X,  Y+1/2,  -Z                         '
                              b'                                        ')

    def test_indexed_extended_header_from_FEI1_file(self):
        with self.newmrc(self.fei1_ext_header_mrc_name) as mrc:
            # FEI1 means use the fei format
            assert mrc.header['exttyp'] == b'FEI1'
            assert mrc.header.nversion == 20140
            assert mrc.header.nsymbt == 786432
            assert mrc.extended_header.nbytes == 786432
            ext = mrc.indexed_extended_header
            assert ext.dtype.kind == 'V'
            # Most fields (e.g. Metadata size) are little-endian in this file
            assert ext.dtype['Metadata size'] == np.dtype('<i4')
            # Bitmasks should always be little-endian
            assert ext.dtype['Bitmask 1'] == np.dtype('<u4')
            assert ext.dtype['Microscope type'] == np.dtype('|S16')
            assert ext[0]['Metadata size'] == 768
            assert ext[0]['Metadata version'] == 0
            assert ext[0]['Microscope type'] == b'TITAN52336320'
            assert ext[0]['HT'] == 300000.0

    def test_indexed_extended_header_from_FEI2_file(self):
        with self.newmrc(self.fei2_ext_header_mrc_name) as mrc:
            # FEI2 means use the fei format
            assert mrc.header['exttyp'] == b'FEI2'
            assert mrc.header.nversion == 20140
            assert mrc.header.nsymbt == 909312
            assert mrc.extended_header.nbytes == 909312
            ext = mrc.indexed_extended_header
            assert ext.dtype.kind == 'V'
            # Most fields (e.g. Metadata size) are little-endian in this file
            assert ext.dtype['Metadata size'] == np.dtype('<i4')
            # Bitmasks should always be little-endian
            assert ext.dtype['Bitmask 1'] == np.dtype('<u4')
            assert ext.dtype['Microscope type'] == np.dtype('|S16')
            assert ext[0]['Metadata size'] == 888
            assert ext[0]['Metadata version'] == 2
            assert ext[0]['Microscope type'] == b'TITAN52337720'
            assert ext[0]['HT'] == 300000.0
            assert ext[0]['Scan rotation'] == 0.0
            assert ext[0]['Detector commercial name'] == b''

    def test_cannot_edit_extended_header_in_read_only_mode(self):
        with self.newmrc(self.ext_header_mrc_name, mode='r') as mrc:
            assert not mrc.extended_header.flags.writeable
            with self.assertRaisesRegex(ValueError, 'read-only'):
                mrc.extended_header.fill(b'a')

    def test_cannot_set_extended_header_in_read_only_mode(self):
        with self.newmrc(self.example_mrc_name, mode='r') as mrc:
            assert not mrc.extended_header.flags.writeable
            with self.assertRaisesRegex(ValueError, 'read-only'):
                mrc.set_extended_header(np.zeros(5))

    def test_voxel_size_is_read_correctly(self):
        with self.newmrc(self.example_mrc_name) as mrc:
            self.assertAlmostEqual(mrc.voxel_size.x, 11.400000, places=6)
            self.assertAlmostEqual(mrc.voxel_size.y, 11.400000, places=6)
            self.assertAlmostEqual(mrc.voxel_size.z, 11.400000, places=6)

    def test_stream_can_be_read_again(self):
        with self.newmrc(self.example_mrc_name) as mrc:
            orig_data = mrc.data.copy()
            mrc._read()
            np.testing.assert_array_equal(orig_data, mrc.data)

    @unittest.skipIf(pathlib_unavailable, "pathlib not available")
    def test_opening_with_pathlib(self):
        path = Path(self.example_mrc_name)
        with self.newmrc(path) as mrc:
            assert self.example_mrc_name in repr(mrc)

    ############################################################################
    #
    # Tests which do not depend on any existing files
    #

    def test_opening_nonexistent_file(self):
        with self.assertRaisesRegex(Exception, "No such file"):
            self.newmrc('no_file')

    def test_opening_file_with_unknown_mode(self):
        with self.newmrc(self.temp_mrc_name, mode='w+') as mrc:
            mrc.header.mode = 10
        with self.assertRaisesRegex(ValueError, "Unrecognised mode"):
            self.newmrc(self.temp_mrc_name)

    def test_can_read_and_flush_stream_repeatedly(self):
        orig_data = np.arange(12, dtype=np.int16).reshape(3, 4)
        with self.newmrc(self.temp_mrc_name, mode='w+') as mrc:
            mrc.set_data(orig_data.copy())
            mrc.flush()
            np.testing.assert_array_equal(orig_data, mrc.data)
            mrc._read()
            np.testing.assert_array_equal(orig_data, mrc.data)
            mrc._read()
            mrc.flush()
            mrc.flush()
            mrc._read()
            mrc._read()
            mrc.flush()
            np.testing.assert_array_equal(orig_data, mrc.data)

    def test_cannot_use_invalid_file_modes(self):
        for mode in ('w', 'a', 'a+'):
            with self.assertRaisesRegex(ValueError, "Mode '.+' not supported"):
                self.newmrc(self.temp_mrc_name, mode=mode)

    def test_cannot_accidentally_overwrite_file(self):
        assert not os.path.exists(self.temp_mrc_name)
        open(self.temp_mrc_name, 'w+').close()
        assert os.path.exists(self.temp_mrc_name)
        with self.assertRaisesRegex(ValueError, "already exists"):
            self.newmrc(self.temp_mrc_name, mode='w+')

    def test_can_deliberately_overwrite_file(self):
        assert not os.path.exists(self.temp_mrc_name)
        open(self.temp_mrc_name, 'w+').close()
        assert os.path.exists(self.temp_mrc_name)
        self.newmrc(self.temp_mrc_name, mode='w+', overwrite=True).close()

    def test_warning_issued_if_file_is_too_large(self):
        with self.newmrc(self.temp_mrc_name, mode='w+') as mrc:
            mrc.set_data(np.arange(12, dtype=np.int16).reshape(3, 4))
            # Call internal _set_new_data() method to add an extra row of data
            # without updating the header
            mrc._set_new_data(np.arange(16, dtype=np.int16).reshape(4, 4))
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            self.newmrc(self.temp_mrc_name)
            assert len(w) == 1
            assert issubclass(w[0].category, RuntimeWarning)
            assert "file is 8 bytes larger than expected" in str(w[0].message)

    def test_exception_raised_if_file_is_too_small_for_reading_extended_header(self):
        with self.newmrc(self.temp_mrc_name, mode='w+') as mrc:
            mrc.set_data(np.arange(24, dtype=np.int16).reshape(2, 3, 4))
            mrc.header.nsymbt = 49
        expected_error_msg = ("Expected 49 bytes in extended header "
                              "but could only read 48")
        with self.assertRaisesRegex(ValueError, expected_error_msg):
            self.newmrc(self.temp_mrc_name)

    def test_exception_raised_if_file_is_too_small_for_reading_data(self):
        with self.newmrc(self.temp_mrc_name, mode='w+') as mrc:
            mrc.set_data(np.arange(24, dtype=np.int16).reshape(2, 3, 4))
            assert mrc.header.mz == 2
            mrc.header.mz = mrc.header.nz = 3
        expected_error_msg = ("Expected 72 bytes in data block"
                              " but limit is 48")
        with self.assertRaisesRegex(ValueError, expected_error_msg):
            self.newmrc(self.temp_mrc_name)

    @unittest.skipIf(sys.maxsize <= np.iinfo(np.int32).max, "can't run test on 32-bit")
    def test_data_is_not_read_if_dimensions_are_too_huge(self):
        # Prepare x, y and z counts to try to trigger an out-of-memory error
        # The individual values need to fit in int32 values to be stored in the header
        # and their product must be as large as possible while still less than
        # sys.maxsize (if larger, it triggers an index overflow error instead)
        max_i4 = np.iinfo('i4').max
        max_arr = sys.maxsize
        nx = max_i4
        ny = min(max_i4, max_arr // nx)
        nz = min(max_i4, max_arr // (nx * ny))
        # Check that an allocation of this size really does cause a memory error
        with self.assertRaises(MemoryError):
            _ = bytearray(nx * ny * nz)

        # Now put these values into a file
        with self.newmrc(self.temp_mrc_name, mode='w+') as mrc:
            mrc.set_data(np.arange(24, dtype=np.int8).reshape(2, 3, 4))
            mrc.header.mx = mrc.header.nx = nx
            mrc.header.my = mrc.header.ny = ny
            mrc.header.mz = mrc.header.nz = nz

        # And now check that if we open the file, we avoid the problem and don't try
        # to allocate enough memory to cause an out-of-memory error
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            with self.newmrc(self.temp_mrc_name, permissive=True) as mrc:
                assert mrc.data is None
            assert len(w) > 0
            assert str(w[0].message) in (
                "Expected {} bytes in data block but limit is 24".format(nx * ny * nz),
                "Error opening memmap",
            )
            assert issubclass(w[0].category, RuntimeWarning)
            assert len(w) == 1

    def test_can_edit_header_in_read_write_mode(self):
        with self.newmrc(self.temp_mrc_name, mode='w+') as mrc:
            mrc.set_data(np.arange(12, dtype=np.int16).reshape(3, 4))
        with self.newmrc(self.temp_mrc_name, mode='r+') as mrc:
            assert mrc.header.ispg == 0
            assert mrc.header.flags.writeable
            mrc.header.ispg = 1
            assert mrc.header.ispg == 1

    def test_cannot_edit_header_in_read_only_mode(self):
        with self.newmrc(self.temp_mrc_name, mode='w+') as mrc:
            mrc.set_data(np.arange(12, dtype=np.int16).reshape(3, 4))
        with self.newmrc(self.temp_mrc_name, mode='r') as mrc:
            assert mrc.header.ispg == 0
            assert not mrc.header.flags.writeable
            with self.assertRaisesRegex(ValueError, 'read-only'):
                mrc.header.ispg = 1
        with self.newmrc(self.temp_mrc_name, mode='r') as mrc:
            assert mrc.header.ispg == 0

    def test_creating_extended_header(self):
        data = np.arange(12, dtype=np.int16).reshape(3, 4)
        extended_header = np.array('example extended header', dtype='S')
        with self.newmrc(self.temp_mrc_name, mode='w+') as mrc:
            mrc.set_data(data)
            mrc.set_extended_header(extended_header)
            np.testing.assert_array_equal(mrc.data, data)
        with self.newmrc(self.temp_mrc_name, mode='r') as mrc:
            # Change the extended header dtype to a string for comparison
            mrc.extended_header.dtype = 'S{}'.format(mrc.extended_header.nbytes)
            np.testing.assert_array_equal(mrc.extended_header, extended_header)
            np.testing.assert_array_equal(mrc.data, data)

    def test_removing_extended_header(self):
        data = np.arange(12, dtype=np.int16).reshape(3, 4)
        extended_header = np.array('example extended header', dtype='S')
        with self.newmrc(self.temp_mrc_name, mode='w+') as mrc:
            mrc.set_data(data)
            mrc.set_extended_header(extended_header)
        with self.newmrc(self.temp_mrc_name, mode='r+') as mrc:
            mrc.set_extended_header(np.array(()))
            mrc.flush()
            assert mrc.header.nsymbt == 0
            file_size = mrc._iostream.tell() # relies on flush() leaving stream at end
            assert file_size == mrc.header.nbytes + mrc.data.nbytes

    def test_extended_header_with_incorrect_type(self):
        data = np.arange(12, dtype=np.int16).reshape(3, 4)
        extended_header = np.array('example extended header', dtype='S')
        with self.newmrc(self.temp_mrc_name, mode='w+') as mrc:
            mrc.set_data(data)
            mrc.set_extended_header(extended_header)
            mrc.header.exttyp = b'FEI1'
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            with self.newmrc(self.temp_mrc_name, mode='r+') as mrc:
                # Test that the file is still read, and the dtype falls back to 'V'
                assert mrc.extended_header.dtype.kind == 'V'
                assert mrc.indexed_extended_header is None
                mrc.extended_header.dtype = 'S{}'.format(mrc.extended_header.nbytes)
                np.testing.assert_array_equal(mrc.extended_header, extended_header)
            assert len(w) == 1
            assert "FEI1" in str(w[0].message)
            assert "extended header" in str(w[0].message)

    def test_can_edit_data_in_read_write_mode(self):
        with self.newmrc(self.temp_mrc_name, mode='w+') as mrc:
            mrc.set_data(np.arange(12, dtype=np.int16).reshape(3, 4))
        with self.newmrc(self.temp_mrc_name, mode='r+') as mrc:
            assert mrc.data[1,1] == 5
            assert mrc.data.flags.writeable
            mrc.data[1,1] = 0
            assert mrc.data[1,1] == 0

    def test_cannot_edit_data_in_read_only_mode(self):
        with self.newmrc(self.temp_mrc_name, mode='w+') as mrc:
            mrc.set_data(np.arange(12, dtype=np.int16).reshape(3, 4))
        with self.newmrc(self.temp_mrc_name, mode='r') as mrc:
            assert mrc.data[1,1] == 5
            assert not mrc.data.flags.writeable
            with self.assertRaisesRegex(ValueError, 'read-only'):
                mrc.data[1,1] = 0

    def test_header_only_mode_does_not_read_data(self):
        with self.newmrc(self.temp_mrc_name, mode='w+') as mrc:
            mrc.set_data(np.arange(12, dtype=np.int16).reshape(3, 4))
        with self.newmrc(self.temp_mrc_name, mode='r', header_only=True) as mrc:
            assert mrc.header is not None
            assert mrc.extended_header is not None
            assert mrc.data is None

    def test_writing_image_mode_0(self):
        x, y = 10, 9
        data = np.linspace(-128, 127, x * y, dtype=np.int8).reshape(y, x)
        name = os.path.join(self.test_output, 'test_img_10x9_mode0.mrc')

        # Write data
        with self.newmrc(name, mode='w+') as mrc:
            mrc.set_data(data)

        # Re-read data and check header and data values
        with self.newmrc(name) as mrc:
            np.testing.assert_array_equal(mrc.data, data)
            assert not mrc.is_image_stack()
            assert mrc.header.ispg == IMAGE_STACK_SPACEGROUP
            assert mrc.header.nx == mrc.header.mx == x
            assert mrc.header.ny == mrc.header.my == y
            assert mrc.header.nz == mrc.header.mz == 1

    def test_writing_image_unsigned_bytes(self):
        x, y = 10, 9
        data = np.linspace(0, 255, x * y, dtype=np.uint8).reshape(y, x)
        name = os.path.join(self.test_output, 'test_img_10x9_uint8.mrc')

        # Write data
        with self.newmrc(name, mode='w+') as mrc:
            mrc.set_data(data)

            # Check data has been converted to mode 6
            np.testing.assert_array_equal(mrc.data, data)
            assert mrc.header.mode == 6
            assert mrc.data.dtype == np.uint16

    def write_file_then_read_and_assert_data_unchanged(self, name, data):
        with self.newmrc(name, mode='w+') as mrc:
            mrc.set_data(data)
        with self.newmrc(name) as mrc:
            np.testing.assert_array_equal(mrc.data, data)
            assert mrc.data.dtype == data.dtype

    def test_writing_image_mode_1_native_byte_order(self):
        data = np.linspace(-32768, 32767, 90, dtype=np.int16).reshape(9, 10)
        name = os.path.join(self.test_output, 'test_img_10x9_mode1_native.mrc')
        self.write_file_then_read_and_assert_data_unchanged(name, data)

    def test_writing_image_mode_1_little_endian(self):
        data = np.linspace(-32768, 32767, 90, dtype='<i2').reshape(9, 10)
        name = os.path.join(self.test_output, 'test_img_10x9_mode1_le.mrc')
        self.write_file_then_read_and_assert_data_unchanged(name, data)

    def test_writing_image_mode_1_big_endian(self):
        data = np.linspace(-32768, 32767, 90, dtype='>i2').reshape(9, 10)
        name = os.path.join(self.test_output, 'test_img_10x9_mode1_be.mrc')
        self.write_file_then_read_and_assert_data_unchanged(name, data)

    def test_writing_image_mode_2_native_byte_order(self):
        data = create_test_float32_array()
        name = os.path.join(self.test_output, 'test_img_10x9_mode2_native.mrc')
        self.write_file_then_read_and_assert_data_unchanged(name, data)

    def test_writing_image_mode_2_little_endian(self):
        data = create_test_float32_array(np.dtype('<f4'))
        name = os.path.join(self.test_output, 'test_img_10x9_mode2_le.mrc')
        self.write_file_then_read_and_assert_data_unchanged(name, data)

    def test_writing_image_mode_2_big_endian(self):
        data = create_test_float32_array(np.dtype('>f4'))
        name = os.path.join(self.test_output, 'test_img_10x9_mode2_be.mrc')
        self.write_file_then_read_and_assert_data_unchanged(name, data)

    def test_writing_image_mode_2_with_inf_and_nan(self):
        # Make an array of test data
        data = create_test_float32_array()

        # Set some unusual values
        data[4][0] = np.nan
        data[4][1] = np.inf
        data[4][2] = -np.inf

        # Write the data to a file and test it's read back correctly
        name = os.path.join(self.test_output, 'test_img_10x9_mode2_inf_nan.mrc')
        # Suppress warnings from statistics calculations with inf and nan
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", RuntimeWarning)
            self.write_file_then_read_and_assert_data_unchanged(name, data)

    def test_writing_image_float16(self):
        x, y = 10, 9
        data = np.linspace(-65504, 65504, x * y, dtype=np.float16).reshape(y, x)
        name = os.path.join(self.test_output, 'test_img_10x9_float16.mrc')

        # Write data
        with self.newmrc(name, mode='w+') as mrc:
            mrc.set_data(data)

            # Check data has been converted to mode 2
            np.testing.assert_array_equal(mrc.data, data)
            assert mrc.header.mode == 12
            assert mrc.data.dtype == np.float16

    def test_writing_image_mode_4_native_byte_order(self):
        data = create_test_complex64_array()
        name = os.path.join(self.test_output, 'test_img_10x9_mode4_native.mrc')
        # Suppress complex casting warnings from statistics calculations
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", ComplexWarning)
            self.write_file_then_read_and_assert_data_unchanged(name, data)

    def test_writing_image_mode_4_little_endian(self):
        data = create_test_complex64_array().astype('<c8')
        name = os.path.join(self.test_output, 'test_img_10x9_mode4_le.mrc')
        # Suppress complex casting warnings from statistics calculations
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", ComplexWarning)
            self.write_file_then_read_and_assert_data_unchanged(name, data)

    def test_writing_image_mode_4_big_endian(self):
        data = create_test_complex64_array().astype('>c8')
        name = os.path.join(self.test_output, 'test_img_10x9_mode4_be.mrc')
        # Suppress complex casting warnings from statistics calculations
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", ComplexWarning)
            self.write_file_then_read_and_assert_data_unchanged(name, data)

    def test_writing_image_mode_4_with_inf_and_nan(self):
        # Make an array of test data
        data = create_test_complex64_array()

        # Set some unusual values
        data[4][0] = (0 + 0j) * np.nan   # =(nan+nan*j)
        data[4][1] = (1 + 1j) * np.inf   # =(inf+inf*j)
        data[4][2] = (-1 - 1j) * np.inf  # =(-inf-inf*j)
        data[4][3] = (1 - 1j) * np.inf   # =(inf-inf*j)
        data[4][4] = (-1 + 1j) * np.inf  # =(-inf+inf*j)

        # Write the data to a file and test it's read back correctly
        name = os.path.join(self.test_output, 'test_img_10x9_mode4_inf_nan.mrc')
        # Suppress warnings from statistics calculations with inf and nan
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", RuntimeWarning)
            self.write_file_then_read_and_assert_data_unchanged(name, data)

    def test_writing_image_mode_6_native_byte_order(self):
        data = np.linspace(0, 65535, 90, dtype=np.int16).reshape(9, 10)
        name = os.path.join(self.test_output, 'test_img_10x9_mode6_native.mrc')
        self.write_file_then_read_and_assert_data_unchanged(name, data)

    def test_writing_image_mode_6_little_endian(self):
        data = np.linspace(0, 65535, 90, dtype='<u2').reshape(9, 10)
        name = os.path.join(self.test_output, 'test_img_10x9_mode6_le.mrc')
        self.write_file_then_read_and_assert_data_unchanged(name, data)

    def test_writing_image_mode_6_big_endian(self):
        data = np.linspace(0, 65535, 90, dtype='>u2').reshape(9, 10)
        name = os.path.join(self.test_output, 'test_img_10x9_mode6_be.mrc')
        self.write_file_then_read_and_assert_data_unchanged(name, data)

    def test_writing_image_stack_mode_2_native_byte_order(self):
        x, y, z = 10, 9, 5
        img = np.linspace(-1e6, 1e6, x * y, dtype=np.float32).reshape(y, x)
        stack = np.arange(1, 6, dtype=np.float32).reshape(z, 1, 1) * img
        name = os.path.join(self.test_output, 'test_img_stack_10x9x5_mode2_native.mrc')

        # Write data
        with self.newmrc(name, mode='w+') as mrc:
            mrc.set_data(stack)
            mrc.set_image_stack()

        # Re-read data and check header and data values
        with self.newmrc(name) as mrc:
            np.testing.assert_array_equal(mrc.data, stack)
            assert mrc.is_image_stack()
            assert mrc.header.ispg == IMAGE_STACK_SPACEGROUP
            assert mrc.header.nx == mrc.header.mx == x
            assert mrc.header.ny == mrc.header.my == y
            assert mrc.header.mz == 1
            assert mrc.header.nz == z

    def test_writing_volume_mode_1_native_byte_order(self):
        x, y, z = 10, 9, 5
        img = np.linspace(-32768, 32767, x * y, dtype=np.int16).reshape(y, x)
        vol = img // np.arange(1, 6, dtype=np.int16).reshape(z, 1, 1)
        name = os.path.join(self.test_output, 'test_vol_10x9x5_mode1_native.mrc')

        # Write data
        with self.newmrc(name, mode='w+') as mrc:
            mrc.set_data(vol)

        # Re-read data and check header and data values
        with self.newmrc(name) as mrc:
            np.testing.assert_array_equal(mrc.data, vol)
            assert mrc.header.ispg == VOLUME_SPACEGROUP
            assert mrc.header.nx == mrc.header.mx == x
            assert mrc.header.ny == mrc.header.my == y
            assert mrc.header.mz == mrc.header.nz == z

    def test_writing_volume_stack_mode_1_native_byte_order(self):
        x, y, z, nvol = 10, 9, 5, 3
        img = np.linspace(-32768, 32767, x * y, dtype=np.int16).reshape(y, x)
        vol = img // np.arange(1, 6, dtype=np.int16).reshape(z, 1, 1)
        stack = vol * np.array([-1, 0, 1], dtype=np.int16).reshape(nvol, 1, 1, 1)
        name = os.path.join(self.test_output, 'test_vol_stack_10x9x5x3_mode1_native.mrc')

        # Write data
        with self.newmrc(name, mode='w+') as mrc:
            mrc.set_data(stack)

        # Re-read data and check header and data values
        with self.newmrc(name) as mrc:
            np.testing.assert_array_equal(mrc.data, stack)
            assert mrc.header.ispg == VOLUME_STACK_SPACEGROUP
            assert mrc.header.nx == mrc.header.mx == x
            assert mrc.header.ny == mrc.header.my == y
            assert mrc.header.mz == z
            assert mrc.header.nz == z * nvol

    def test_data_transposed_in_place_is_written_without_errors(self):
        # Quite unlikely that anyone will mess with the data array like this,
        # but still worth making sure the flush() call is robust!
        x, y, z = 10, 9, 5
        img = np.linspace(-32768, 32767, x * y, dtype=np.int16).reshape(y, x)
        vol = img // np.arange(1, 6, dtype=np.int16).reshape(z, 1, 1)
        transposed_vol = vol.transpose()

        # Write data and confirm it's C-contiguous
        mrc = self.newmrc(self.temp_mrc_name, mode='w+')
        mrc.set_data(vol)
        mrc.flush()
        assert mrc.data.flags.c_contiguous == True

        # Transpose the data array in-place
        strides = mrc.data.strides
        mrc.data.shape = mrc.data.shape[::-1]
        mrc.data.strides = strides[::-1]
        # Check this is an effective way to do in-place transpose()
        np.testing.assert_array_equal(transposed_vol, mrc.data)

        # Confirm data is no longer C-contiguous
        assert mrc.data.flags.c_contiguous == False

        # Flush and close should work without errors
        mrc.flush()
        mrc.close()

    def test_transposed_data_is_made_contiguous_on_set(self):
        # Quite unlikely that anyone will mess with the data array like this,
        # but still worth making sure the flush() call is robust!
        x, y, z = 10, 9, 5
        img = np.linspace(-32768, 32767, x * y, dtype=np.int16).reshape(y, x)
        vol = img // np.arange(1, 6, dtype=np.int16).reshape(z, 1, 1)
        vol = vol.transpose()

        assert vol.flags.c_contiguous == False

        # Write data and confirm it's C-contiguous
        with self.newmrc(self.temp_mrc_name, mode='w+') as mrc:
            mrc.set_data(vol)
            assert mrc.data.flags.c_contiguous == True

    def test_permissive_read_with_wrong_machine_stamp(self):
        data = np.arange(12, dtype=np.int16).reshape(3, 4)
        with self.newmrc(self.temp_mrc_name, mode='w+') as mrc:
            mrc.set_data(data)
            wrong_byte_order = mrc.header.mode.dtype.newbyteorder().byteorder
            mrc.header.machst = utils.machine_stamp_from_byte_order(wrong_byte_order)
        with self.assertRaisesRegex(ValueError, "Unrecognised mode"):
            self.newmrc(self.temp_mrc_name)
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            with self.newmrc(self.temp_mrc_name, permissive=True) as mrc:
                np.testing.assert_array_equal(mrc.data, data)
            assert len(w) == 1
            assert "Machine stamp" in str(w[0].message)
            assert "byte order" in str(w[0].message)




def create_test_float32_array(dtype=np.float32):
    """Create a 10 x 9 array of float values.

    Values above 10^19 cause data.std(dtype=np.float32) to overflow, so we
    stay inside that range to avoid warnings."""
    data = np.zeros((9, 10), dtype=dtype)
    data[:4] = np.negative(np.logspace(19, -38, 40).reshape(4, 10))
    data[5:] = np.logspace(-38, 19, 40).reshape(4, 10)
    return data


def create_test_complex64_array():
    floats = create_test_float32_array()
    data = 0.1j * floats[::-1]
    data += floats
    assert data.dtype.type == np.complex64
    return data


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2016, Science and Technology Facilities Council
# This software is distributed under a BSD licence. See LICENSE.txt.
"""
dtypes
------

numpy dtypes used by the ``mrcfile.py`` library.

The dtypes are defined in a separate module because they do not interact nicely
with the ``from __future__ import unicode_literals`` feature used in the rest
of the package.

"""

# Import Python 3 features for future-proofing
# Deliberately do NOT import unicode_literals due to a bug in numpy dtypes:
# https://github.com/numpy/numpy/issues/2407
from __future__ import absolute_import, division, print_function

import numpy as np

from .utils import normalise_byte_order


HEADER_DTYPE = np.dtype([
    ('nx', 'i4'),          # Number of columns
    ('ny', 'i4'),          # Number of rows
    ('nz', 'i4'),          # Number of sections
    
    ('mode', 'i4'),        # Mode; indicates type of values stored in data block
    
    ('nxstart', 'i4'),     # Starting point of sub-image
    ('nystart', 'i4'),
    ('nzstart', 'i4'),
    
    ('mx', 'i4'),          # Grid size in X, Y and Z
    ('my', 'i4'),
    ('mz', 'i4'),
    
    ('cella', [            # Cell size in angstroms
        ('x', 'f4'),
        ('y', 'f4'),
        ('z', 'f4')
    ]),
    
    ('cellb', [            # Cell angles in degrees
        ('alpha', 'f4'),
        ('beta', 'f4'),
        ('gamma', 'f4')
    ]),
    
    ('mapc', 'i4'),        # map column  1=x,2=y,3=z.
    ('mapr', 'i4'),        # map row     1=x,2=y,3=z.
    ('maps', 'i4'),        # map section 1=x,2=y,3=z.
    
    ('dmin', 'f4'),        # Minimum pixel value
    ('dmax', 'f4'),        # Maximum pixel value
    ('dmean', 'f4'),       # Mean pixel value
    
    ('ispg', 'i4'),        # space group number
    ('nsymbt', 'i4'),      # number of bytes in extended header
    
    ('extra1', 'V8'),      # extra space, usage varies by application
    ('exttyp', 'S4'),      # code for the type of extended header
    ('nversion', 'i4'),    # version of the MRC format
    ('extra2', 'V84'),     # extra space, usage varies by application
    
    ('origin', [           # Origin of image
        ('x', 'f4'),
        ('y', 'f4'),
        ('z', 'f4')
    ]),
    
    ('map', 'S4'),         # Contains 'MAP ' to identify file type
    ('machst', 'u1', 4),   # Machine stamp; identifies byte order
    
    ('rms', 'f4'),         # RMS deviation of densities from mean density
    
    ('nlabl', 'i4'),       # Number of labels with useful data
    ('label', 'S80', 10)   # 10 labels of 80 characters
])


VOXEL_SIZE_DTYPE = np.dtype([
    ('x', 'f4'),
    ('y', 'f4'),
    ('z', 'f4')
])


NSTART_DTYPE = np.dtype([
    ('x', 'i4'),
    ('y', 'i4'),
    ('z', 'i4')
])


# FEI extended header dtype for metadata version 0, as described in the EPU
# manual. Note that the FEI documentation is unclear about the endianness of
# the data in the extended header, except for the bitmask fields which are
# specified as little-endian. To construct mixed-endian dtypes, we initially
# construct a big-endian type, with only the bitmasks little-endian, then we
# swap all fields to little-endian if we need a fully little-endian type.
fei_dtype_dict = [
    ('Metadata size', '>i4'),
    ('Metadata version', '>i4'),
    ('Bitmask 1', '<u4'),
    ('Timestamp', '>f8'),  # Not specified, but suspect this is in days after 1/1/1900
    ('Microscope type', 'S16'),
    ('D-Number', 'S16'),
    ('Application', 'S16'),
    ('Application version', 'S16'),
    ('HT', '>f8'),
    ('Dose', '>f8'),
    ('Alpha tilt', '>f8'),
    ('Beta tilt', '>f8'),
    ('X-Stage', '>f8'),
    ('Y-Stage', '>f8'),
    ('Z-Stage', '>f8'),
    ('Tilt axis angle', '>f8'),
    ('Dual axis rotation', '>f8'),
    ('Pixel size X', '>f8'),
    ('Pixel size Y', '>f8'),
    ('Unused range', 'S48'),
    ('Defocus', '>f8'),
    ('STEM Defocus', '>f8'),
    ('Applied defocus', '>f8'),
    ('Instrument mode', '>i4'),
    ('Projection mode', '>i4'),
    ('Objective lens mode', 'S16'),
    ('High magnification mode', 'S16'),
    ('Probe mode', '>i4'),
    ('EFTEM On', '?'),
    ('Magnification', '>f8'),
    ('Bitmask 2', '<u4'),
    ('Camera length', '>f8'),
    ('Spot index', '>i4'),
    ('Illuminated area', '>f8'),
    ('Intensity', '>f8'),
    ('Convergence angle', '>f8'),
    ('Illumination mode', 'S16'),
    ('Wide convergence angle range', '?'),
    ('Slit inserted', '?'),
    ('Slit width', '>f8'),
    ('Acceleration voltage offset', '>f8'),
    ('Drift tube voltage', '>f8'),
    ('Energy shift', '>f8'),
    ('Shift offset X', '>f8'),
    ('Shift offset Y', '>f8'),
    ('Shift X', '>f8'),
    ('Shift Y', '>f8'),
    ('Integration time', '>f8'),
    ('Binning Width', '>i4'),
    ('Binning Height', '>i4'),
    ('Camera name', 'S16'),
    ('Readout area left', '>i4'),
    ('Readout area top', '>i4'),
    ('Readout area right', '>i4'),
    ('Readout area bottom', '>i4'),
    ('Ceta noise reduction', '?'),
    ('Ceta frames summed', '>i4'),
    ('Direct detector electron counting', '?'),
    ('Direct detector align frames', '?'),
    ('Camera param reserved 0', '>i4'),
    ('Camera param reserved 1', '>i4'),
    ('Camera param reserved 2', '>i4'),
    ('Camera param reserved 3', '>i4'),
    ('Bitmask 3', '<u4'),
    ('Camera param reserved 4', '>i4'),
    ('Camera param reserved 5', '>i4'),
    ('Camera param reserved 6', '>i4'),
    ('Camera param reserved 7', '>i4'),
    ('Camera param reserved 8', '>i4'),
    ('Camera param reserved 9', '>i4'),
    ('Phase Plate', '?'),
    ('STEM Detector name', 'S16'),
    ('Gain', '>f8'),
    ('Offset', '>f8'),
    ('STEM param reserved 0', '>i4'),
    ('STEM param reserved 1', '>i4'),
    ('STEM param reserved 2', '>i4'),
    ('STEM param reserved 3', '>i4'),
    ('STEM param reserved 4', '>i4'),
    ('Dwell time', '>f8'),
    ('Frame time', '>f8'),
    ('Scan size left', '>i4'),
    ('Scan size top', '>i4'),
    ('Scan size right', '>i4'),
    ('Scan size bottom', '>i4'),
    ('Full scan FOV X', '>f8'),
    ('Full scan FOV Y', '>f8'),
    ('Element', 'S16'),
    ('Energy interval lower', '>f8'),
    ('Energy interval higher', '>f8'),
    ('Method', '>i4'),
    ('Is dose fraction', '?'),
    ('Fraction number', '>i4'),
    ('Start frame', '>i4'),
    ('End frame', '>i4'),
    ('Input stack filename', 'S80'),
    ('Bitmask 4', '<u4'),
    ('Alpha tilt min', '>f8'),
    ('Alpha tilt max', '>f8')
]
fei1_dtype_big_endian = np.dtype(fei_dtype_dict)


# Additional metadata entries for FEI extended header dtype for metadata version 2,
# as described in https://www.fei-software-center.com/tem-apps/MRC-2014-Specifications/
fei_dtype_dict.extend([
    ('Scan rotation', '>f8'),
    ('Diffraction pattern rotation', '>f8'),
    ('Image rotation', '>f8'),
    ('Scan mode enumeration', '>i4'),
    ('Acquisition time stamp', '>i8'),
    ('Detector commercial name', 'S16'),
    ('Start tilt angle', '>f8'),
    ('End tilt angle', '>f8'),
    ('Tilt per image', '>f8'),
    ('Tilt speed', '>f8'),
    ('Beam center X pixel', '>i4'),
    ('Beam center Y pixel', '>i4'),
    ('CFEG flash timestamp', '>i8'),
    ('Phase plate position index', '>i4'),
    ('Objective aperture name', 'S16')
])
fei2_dtype_big_endian = np.dtype(fei_dtype_dict)


def get_ext_header_dtype(exttyp, byte_order='='):
    """Get a dtype for an extended header.

    Args:
        exttyp: One of ``b'FEI1'`` or ``b'FEI2'``, which are currently the only
            supported extended header types.
        byte_order: One of ``=``, ``<`` or ``>``.

    Returns:
        A :class:`numpy dtype <numpy.dtype>` object for the extended header, or
        :data:`None`

    Raises:
        :exc:`ValueError`: If ``byte_order`` is not one of ``=``, ``<`` or ``>``.
    """
    normalised_byte_order = normalise_byte_order(byte_order)
    dtype = None
    if exttyp == b'FEI1':
        dtype = fei1_dtype_big_endian
    elif exttyp == b'FEI2':
        dtype = fei2_dtype_big_endian
    # If returning a little-endian FEI1 or FEI2 dtype, need to swap the byte order
    if dtype is not None and normalised_byte_order == '<':
        dtype = dtype.newbyteorder('<')
    return dtype



# Copyright (c) 2018, Science and Technology Facilities Council
# This software is distributed under a BSD licence. See LICENSE.txt.

"""
Tests for mrcfile command line entry point functions.
"""

# Import Python 3 features for future-proofing
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import io
import os
import shutil
import sys
import tempfile
import unittest

from mrcfile import command_line, validator
from . import helpers


class CommandLineTest(helpers.AssertRaisesRegexMixin, unittest.TestCase):
    
    """Unit tests for mrcfile command line functions.
    
    """
    
    def setUp(self):
        super(CommandLineTest, self).setUp()
        
        # Set up test files and names to be used
        self.test_data = helpers.get_test_data_path()
        self.test_output = tempfile.mkdtemp()
        
        self.files = [
            os.path.join(self.test_data, 'EMD-3197.map'),
            os.path.join(self.test_data, 'emd_3001.map.gz'),
            os.path.join(self.test_data, 'EMD-3197.map.bz2')
        ]
        
        # Set up stream to catch print output from print_header()
        self.print_stream = io.StringIO()
        
        # Replace stdout and stderr to capture output for checking
        self.orig_stdout = sys.stdout
        self.orig_stderr = sys.stderr
        sys.stdout = io.StringIO()
        sys.stderr = io.StringIO()
    
    def tearDown(self):
        # Restore stdout and stderr
        sys.stdout = self.orig_stdout
        sys.stderr = self.orig_stderr
        
        if os.path.exists(self.test_output):
            shutil.rmtree(self.test_output)
        super(CommandLineTest, self).tearDown()

    def test_print_header_no_args(self):
        command_line.print_headers([], print_file=self.print_stream)
        assert len(self.print_stream.getvalue()) == 0
        assert len(sys.stdout.getvalue()) == 0
        assert len(sys.stderr.getvalue()) == 0

    def test_print_header_nonexistent_file(self):
        with self.assertRaisesRegex(IOError, "No such file"):
            command_line.print_headers(["nonexistent.mrc"],
                                       print_file=self.print_stream)
        assert len(self.print_stream.getvalue()) == 0
        assert len(sys.stdout.getvalue()) == 0
        assert len(sys.stderr.getvalue()) == 0

    def test_print_header(self):
        command_line.print_headers(self.files, print_file=self.print_stream)
        printed = self.print_stream.getvalue()
        assert len(printed) > 0
        for file in self.files:
            assert "MRC header for " + file in printed
        assert "nx" in printed
        assert "machst          : [68 65  0  0]" in printed
        assert "::::EMDATABANK.org::::EMD-3197::::" in printed
        assert len(sys.stdout.getvalue()) == 0
        assert len(sys.stderr.getvalue()) == 0

    def test_validate_no_args(self):
        result = validator.main([])
        assert result == 0
        assert len(self.print_stream.getvalue()) == 0
        assert len(sys.stdout.getvalue()) == 0
        assert len(sys.stderr.getvalue()) == 0
    
    def test_validate(self):
        result = validator.main(self.files)
        assert result == 1
        stdout = str(sys.stdout.getvalue())
        assert stdout == (
            "Checking if " + self.files[0] + " is a valid MRC2014 file...\n"
            "File does not declare MRC format version 20140 or 20141: nversion = 0\n"
            "Checking if " + self.files[1] + " is a valid MRC2014 file...\n"
            "File does not declare MRC format version 20140 or 20141: nversion = 0\n"
            "Extended header type is undefined or unrecognised: exttyp = ''\n"
            "Checking if " + self.files[2] + " is a valid MRC2014 file...\n"
            "File does not declare MRC format version 20140 or 20141: nversion = 0\n"
        )
        assert len(sys.stderr.getvalue()) == 0


if __name__ == '__main__':
    unittest.main()
